from datetime import timedelta
import datetime
import xarray as xr
import pandas as pd
import tempfile
import os
import numpy as np
import json

from hydrodatasource.configs.config import (
    FS,
    GRID_INTERIM_BUCKET,
    LOCAL_DATA_PATH,
    MC,
    RO,
)
from hydrodatasource.utils.utils import generate_time_intervals
from hydrodatasource.processor.mask import gen_single_mask
from hydrodatasource.reader import minio


def process_data(initial_date, initial_time, mask, dataset):
    # Convert the initial date and time to a datetime object
    current_datetime = datetime.datetime.strptime(
        f"{initial_date} {initial_time}", "%Y-%m-%d %H"
    )
    initial_datetime = datetime.datetime.strptime(
        f"{initial_date} {initial_time}", "%Y-%m-%d %H"
    )
    gfs_reader = minio.GFSReader()
    box = (
        mask.coords["lon"][0] - 0.2,
        mask.coords["lat"][0] - 0.2,
        mask.coords["lon"][-1] + 0.2,
        mask.coords["lat"][-1] + 0.2,
    )

    while True:
        # Your existing code to process the data
        data = gfs_reader.open_dataset(
            creation_date=np.datetime64(current_datetime.strftime("%Y-%m-%d")),
            creation_time=current_datetime.strftime("%H"),
            bbox=box,
            dataset=dataset,
            time_chunks=24,
        )
        data = data.load()
        data = data.to_dataset()
        data = data.transpose("time", "valid_time", "lon", "lat")

        # Adjust the slicing based on the current datetime
        total_hours_passed = int(
            (initial_datetime - current_datetime).total_seconds() / 3600
        )

        # Adjust the slicing based on the total hours passed
        start_slice = total_hours_passed + 1
        end_slice = start_slice + 6
        # Minio上存储的gfs数据都是该时刻未来120小时后的数据，因此这里设置了一个报错
        if start_slice > 120:
            raise Exception("This datetime does not have enough data to merge")
        data = data.isel(time=0, valid_time=slice(start_slice, end_slice))

        # Check for negative values
        data_det = data["total_precipitation_surface"].values
        has_negative = data_det < 0
        if has_negative.any():
            # If negative, roll back time by 6 hours and repeat
            current_datetime -= timedelta(hours=6)
        else:
            # No negative values found, return the data
            return data


def make_gpm_dataset(
    time_periods,
    dataset,
    mask,
):
    gpm_reader = minio.GPMReader()
    latest_data = xr.Dataset()
    box = (
        mask.coords["lon"][0],
        mask.coords["lat"][0],
        mask.coords["lon"][-1],
        mask.coords["lat"][-1],
    )
    for time_num in time_periods:
        start_time = np.datetime64(time_num[0])
        end_time = np.datetime64(time_num[1])
        data = gpm_reader.open_dataset(
            start_time=start_time,
            end_time=end_time,
            dataset=dataset,
            bbox=box,
            time_resolution="30m",
            time_chunks=24,
        )
        data = data.load()
        # data = data.to_dataset()
        # 转换时间维度至Pandas的DateTime格式并创建分组标签
        times = pd.to_datetime(data["time"].values)
        group_labels = times.floor("h")

        # 创建一个新的DataArray，其时间维度与原始数据集匹配
        group_labels_da = xr.DataArray(
            group_labels, coords={"time": data["time"]}, dims=["time"]
        )

        # 对数据进行分组并求和
        merge_gpm_data = data.groupby(group_labels_da).mean("time")

        # 将维度名称从group_labels_da重新命名为'time'
        merge_gpm_data = merge_gpm_data.rename({"group": "time"})
        merge_gpm_data.name = "tp"
        merge_gpm_data = merge_gpm_data.to_dataset()
        w_data = mask["w"]
        w_data_interpolated = w_data.interp(
            lat=merge_gpm_data.lat, lon=merge_gpm_data.lon, method="nearest"
        ).fillna(0)
        w_data_broadcasted = w_data_interpolated.broadcast_like(merge_gpm_data["tp"])
        merge_gpm_data = merge_gpm_data["tp"] * w_data_broadcasted
        if isinstance(latest_data, xr.DataArray):
            latest_data.name = "tp"
            latest_data = latest_data.to_dataset()
        if isinstance(merge_gpm_data, xr.DataArray):
            merge_gpm_data.name = "tp"
            merge_gpm_data = merge_gpm_data.to_dataset()
        is_empty = not latest_data.data_vars
        if is_empty:
            latest_data = merge_gpm_data
        else:
            latest_data = xr.concat([latest_data, merge_gpm_data], dim="time")
    return latest_data


def make_gfs_dataset(
    time_periods,
    dataset,
    mask,
):
    final_latest_data = xr.Dataset()
    for time_num in time_periods:
        start_time = datetime.datetime.strptime(time_num[0], "%Y-%m-%dT%H:%M:%S")
        end_time = datetime.datetime.strptime(time_num[1], "%Y-%m-%dT%H:%M:%S")
        gfs_time_list = generate_time_intervals(start_time, end_time)
        w_data = mask["w"]
        latest_data = xr.Dataset()

        for date, time in gfs_time_list:
            data = process_data(date, time, mask, dataset)
            if isinstance(data, xr.DataArray):
                data = data.to_dataset()
            if "total_precipitation_surface" in data.data_vars:
                data = data.rename({"total_precipitation_surface": "tp"})
            data = data.drop_vars("time")
            # print(data)
            data = data.rename({"valid_time": "time"})
            w_data_interpolated = w_data.interp(
                lat=data.lat, lon=data.lon, method="nearest"
            ).fillna(0)

            # 将 w 数据广播到与当前数据集相同的时间维度上
            w_data_broadcasted = w_data_interpolated.broadcast_like(data["tp"])
            data["tp"] = data["tp"] * w_data_broadcasted
            if isinstance(latest_data, xr.DataArray):
                latest_data.name = "tp"
                latest_data = latest_data.to_dataset()
            if isinstance(data, xr.DataArray):
                data.name = "tp"
                data = data.to_dataset()
            is_empty = not latest_data.data_vars
            latest_data = (
                data if is_empty else xr.concat([latest_data, data], dim="time")
            )
        if isinstance(latest_data, xr.DataArray):
            latest_data.name = "tp"
            latest_data = latest_data.to_dataset()
        if isinstance(final_latest_data, xr.DataArray):
            final_latest_data.name = "tp"
            final_latest_data = final_latest_data.to_dataset()
        is_final_empty = not final_latest_data.data_vars
        if is_final_empty:
            final_latest_data = latest_data
        else:
            final_latest_data = xr.concat([final_latest_data, latest_data], dim="time")
    return final_latest_data


def make_merge_dataset(
    gpm_data, gfs_data, time_periods, gpm_length, gfs_length, time_now_length
):
    # 指定时间段
    for time_num in time_periods:
        start_time = pd.to_datetime(time_num[0])
        end_time = pd.to_datetime(time_num[1])
    time_range = pd.date_range(start=start_time, end=end_time, freq="H")

    # 创建一个空的 xarray 数据集来存储结果
    combined_data = xr.Dataset()

    # 循环处理每个小时的数据
    for specified_time in time_range:
        m_hours = gpm_length
        n_hours = gfs_length

        # 选取specified_time前指定时间段的时间的gpm数据
        gpm_data_filtered = gpm_data.sel(
            time=slice(specified_time - pd.Timedelta(hours=m_hours), specified_time)
        )
        # 选取specified_time后指定时间段的时间的gfs数据
        gfs_data_filtered = gfs_data.sel(
            time=slice(
                specified_time + pd.Timedelta(1),  # gfs数据要在
                specified_time + pd.Timedelta(hours=n_hours),
            )
        )

        gfs_data_interpolated = gfs_data_filtered.interp(
            lat=gpm_data.lat, lon=gpm_data.lon, method="linear"
        )
        combined_hourly_data = xr.concat(
            [gpm_data_filtered, gfs_data_interpolated], dim="time"
        )
        combined_hourly_data = combined_hourly_data.rename({"time": "step"})
        combined_hourly_data["step"] = np.arange(len(combined_hourly_data.step))
        time_now_hour = (
            specified_time
            - pd.Timedelta(hours=m_hours)
            + pd.Timedelta(hours=time_now_length)
        )
        combined_hourly_data.coords["time_now"] = time_now_hour
        combined_hourly_data = combined_hourly_data.expand_dims("time_now")

        # 合并到结果数据集中
        combined_data = xr.merge(
            [combined_data, combined_hourly_data], combine_attrs="override"
        )
    return combined_data


def make1nc41basin(
    basin_id="1_02051500",
    dataname="gpm",
    local_path=LOCAL_DATA_PATH,
    mask_path=os.path.join(LOCAL_DATA_PATH, "dataset-origin", "mask"),
    shp_path=os.path.join(LOCAL_DATA_PATH, "dataset-origin", "shp"),
    dataset="camels",
    time_periods=[["2017-01-01T00:00:00", "2017-01-31T00:00:00"]],
    local_save=True,
    minio_upload=False,
    gpm_length=None,
    gfs_length=None,
    time_now_length=None,
    gpm_path=None,
    gfs_path=None,
):
    if dataset not in ["camels", "wis"]:
        # 流域是国内还是国外，国外是camels，国内是wis
        raise ValueError("Invalid dataset")

    mask = gen_single_mask(basin_id, shp_path, dataname, mask_path)

    # 如果是合并数据，需要额外处理
    if dataname == "merge":
        mask = gen_single_mask(basin_id, shp_path, "gpm", mask_path)
        gpm_data = (
            make_gpm_dataset(time_periods, dataset, mask)
            if gpm_path is None
            else xr.open_dataset(gpm_path)
        )
        mask = gen_single_mask(basin_id, shp_path, "gfs", mask_path)
        gfs_data = (
            make_gfs_dataset(time_periods, dataset, mask)
            if gfs_path is None
            else xr.open_dataset(gfs_path)
        )

    data_functions = {
        "gpm": lambda: make_gpm_dataset(time_periods, dataset, mask),
        "gfs": lambda: make_gfs_dataset(time_periods, dataset, mask),
        "merge": lambda: make_merge_dataset(
            gpm_data, gfs_data, time_periods, gpm_length, gfs_length, time_now_length
        ),
    }

    # 检查数据名称是否有效
    if dataname not in data_functions:
        raise NotImplementedError(
            f"This type of data ({dataname}) is not available for now, please try gpm, gfs, or merge"
        )

    # 根据数据名称调用相应的函数
    latest_data = data_functions[dataname]()

    if local_save:
        local_save_path = os.path.join(local_path, "datasets-interim", basin_id)
        if not os.path.exists(local_save_path):
            os.makedirs(local_save_path)
        local_file_path = os.path.join(
            local_path, "datasets-interim", basin_id, dataname
        )
        local_file_name = local_file_path + ".nc"
        latest_data.to_netcdf(local_file_name)

    if minio_upload:
        object_name = basin_id + "/" + dataname + ".nc"

        # 元数据
        time_periods_str = json.dumps(time_periods)
        metadata = {
            "X-Amz-Meta-Time_Periods": time_periods_str,
        }

        with tempfile.NamedTemporaryFile() as tmp:
            # 将数据集保存到临时文件
            latest_data.to_netcdf(tmp.name)

            # 重置文件读取指针
            tmp.seek(0)

            # 上传到 MinIO
            MC.fput_object(
                GRID_INTERIM_BUCKET, object_name, tmp.name, metadata=metadata
            )

    return latest_data



"""
Author: Tianxu Liu
Date: 2024-03-16 15:55:22
LastEditTime: 2024-07-06 19:08:11
LastEditors: Wenyu Ouyang
Description: test process inq
FilePath: /hydrodatasource/tests/test_process_inq.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import pytest

from hydrodatasource.cleaner.cleaner import Cleaner


@pytest.fixture
def cleaner():
    return Cleaner(
        file_path="/home/liutianxv/sample_data.csv",
        column_id="ID",
        ID_list=None,
        column_flow="INQ",
        column_time="TM",
        start_time=None,
        end_time=None,
        preprocess=True,
        method="kalman",
        save_path="/home/liutianxv/sample_data.csv",
        plot=True,
        window_size=20,
        cutoff_frequency=0.035,
        iterations=3,
        sampling_rate=1.0,
        time_step=1,
        order=5,
        cwt_row=10,
    )


def test_process_inq(cleaner):
    cleaner.process_inq()



"""
Author: Wenyu Ouyang
Date: 2023-10-06 21:17:17
LastEditTime: 2024-03-28 08:22:54
LastEditors: Wenyu Ouyang
Description: Test reading according to AOI
FilePath: \hydrodatasource\tests\test_aoi_reading.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import pytest
import geopandas as gpd
import numpy as np

from hydrodatasource.reader.minio import ERA5LReader, GPMReader, GFSReader


@pytest.fixture()
def geo_file():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(current_dir, "test.geojson")


@pytest.fixture()
def aoi_shp(geo_file):
    return gpd.read_file(geo_file)


@pytest.fixture()
def start_time():
    return np.datetime64("2021-06-01T00:00:00.000000000")


@pytest.fixture()
def end_time():
    return np.datetime64("2021-06-30T23:00:00.000000000")


@pytest.fixture()
def creation_date():
    return np.datetime64("2021-06-01")


def test_read_era5(aoi_shp, start_time, end_time):
    era5 = ERA5LReader()
    ds1 = era5.from_aoi(
        data_variables=["Total precipitation"],
        start_time=start_time,
        end_time=end_time,
        aoi=aoi_shp,
    )
    print(ds1)


def test_read_gpm(aoi_shp, start_time, end_time):
    gpm = GPMReader()
    ds2 = gpm.from_aoi(start_time=start_time, end_time=end_time, aoi=aoi_shp)
    print(ds2)


def test_read_gfs(aoi_shp, creation_date):
    gfs = GFSReader()
    ds3 = gfs.from_aoi(creation_date=creation_date, creation_time="00", aoi=aoi_shp)
    print(ds3)



"""
Author: Shuolong Xu
Date: 2024-02-15 16:40:34
LastEditTime: 2024-03-28 08:42:05
LastEditors: Wenyu Ouyang
Description: Test cases for gpm and gfs data
FilePath: \hydrodatasource\tests\test_gpm_gfs.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
from datetime import datetime

from hydrodatasource.processor.mask import gen_single_mask
from hydrodatasource.configs.config import LOCAL_DATA_PATH
from hydrodatasource.processor.gpm_gfs import make1nc41basin
from hydrodatasource.utils.utils import generate_time_intervals


def test_gen_mask():
    mask = gen_single_mask(
        basin_id="1_02051500",
        dataname="gpm",
    )
    assert mask is not None
    return mask


def test_gen_mask_minio():
    mask = gen_single_mask(
        basin_id="10310500",
        dataname="gpm",
        mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
        minio=True,
    )
    assert mask is not None
    return mask


def test_time_intervals():
    time = generate_time_intervals(
        start_date=datetime(2017, 1, 1), end_date=datetime(2017, 1, 3)
    )
    assert time is not None
    return time


def test_gpm():
    data = make1nc41basin(
        basin_id="1_02051500",
        dataname="gpm",
        local_path=LOCAL_DATA_PATH,
        mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
        shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
        dataset="camels",
        time_periods=[["2017-01-01T00:00:00", "2017-01-31T00:00:00"]],
        local_save=True,
        minio_upload=False,
    )
    assert data is not None
    return data


def test_gfs():
    data = make1nc41basin(
        basin_id="1_02051500",
        dataname="gfs",
        local_path=LOCAL_DATA_PATH,
        mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
        shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
        dataset="camels",
        time_periods=[["2017-01-01T00:00:00", "2017-01-31T00:00:00"]],
        local_save=True,
        minio_upload=False,
    )
    assert data is not None
    return data


def test_merge_without_local_data():
    data = make1nc41basin(
        basin_id="1_02051500",
        dataname="merge",
        local_path=LOCAL_DATA_PATH,
        mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
        shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
        dataset="camels",
        time_periods=[
            ["2017-01-10T00:00:00", "2017-01-11T00:00:00"],
            ["2017-01-15T00:00:00", "2017-01-20T00:00:00"],
        ],
        local_save=True,
        minio_upload=False,
        gpm_length=169,
        gfs_length=23,
        time_now_length=168,
    )
    assert data is not None
    return data


def test_merge_with_local_data():
    data = make1nc41basin(
        basin_id="1_02051500",
        dataname="merge",
        local_path=LOCAL_DATA_PATH,
        mask_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "mask"),
        shp_path=os.path.join(LOCAL_DATA_PATH, "datasets-origin", "shp"),
        dataset="camels",
        time_periods=[["2017-01-01T00:00:00", "2017-01-31T00:00:00"]],
        local_save=True,
        minio_upload=False,
        gpm_length=169,
        gfs_length=23,
        time_now_length=168,
        gpm_path=os.path.join(
            LOCAL_DATA_PATH, "datasets-interim", "1_02051500", "gpm.nc"
        ),
        gfs_path=os.path.join(
            LOCAL_DATA_PATH, "datasets-interim", "1_02051500", "gfs.nc"
        ),
    )
    assert data is not None
    return data



import numpy as np
import pandas as pd
from pykalman import KalmanFilter


def calculate_esm(QFi, Qi):  # 计算平滑度
    numerator_list = []
    denominator_list = []
    for i in range(len(QFi) - 1):
        numerator = (QFi.values[i + 1] - QFi.values[i]) ** 2
        denominator = (Qi.values[i + 1] - Qi.values[i]) ** 2
        numerator_list.append(numerator)
        denominator_list.append(denominator)
    numerator_total = np.sum(numerator_list)
    denominator_total = np.sum(denominator_list)
    return 1 - numerator_total / denominator_total


def get_moving_average_inq(inq_data_df):
    """
    :param inq_data_df: 入库流量表格，需要有TM（时间）和INQ（入库流量）两列
    :return:
    """
    inq_data = inq_data_df["INQ"]
    inq_data_df["TM"] = pd.to_datetime(inq_data_df["TM"], format="%d/%m/%Y %H:%M")
    # 滑动平均
    window_size = 5
    inq_moving_average = np.convolve(
        inq_data_df["INQ"], np.ones(window_size) / window_size, mode="same"
    )
    # 五点三次
    QF = np.zeros(len(inq_data))
    QF[0] = (
        1
        / 70
        * (
            69 * inq_data[0]
            + 4 * inq_data[1]
            - 6 * inq_data[2]
            + 4 * inq_data[3]
            - inq_data[4]
        )
    )
    QF[1] = (
        1
        / 30
        * (
            2 * inq_data[0]
            + 27 * inq_data[1]
            + 12 * inq_data[2]
            - 8 * inq_data[3]
            + 2 * inq_data[4]
        )
    )
    for i in range(2, len(inq_data) - 2):
        QF[i] = (
            1
            / 35
            * (
                (-3) * inq_data[i - 2]
                + 12 * inq_data[i - 1]
                + 17 * inq_data[i]
                + 12 * inq_data[i + 1]
                - 3 * inq_data[i + 2]
            )
        )
    QF[len(inq_data) - 2] = (
        1
        / 35
        * (
            2 * inq_data[len(inq_data) - 5]
            - 8 * inq_data[len(inq_data) - 4]
            + 12 * inq_data[len(inq_data) - 3]
            + 27 * inq_data[len(inq_data) - 2]
            + 2 * inq_data[len(inq_data) - 1]
        )
    )
    QF[len(inq_data) - 1] = (
        1
        / 70
        * (
            (-1) * inq_data[len(inq_data) - 5]
            + 4 * inq_data[len(inq_data) - 4]
            - 6 * inq_data[len(inq_data) - 3]
            + 4 * inq_data[len(inq_data) - 2]
            + 69 * inq_data[len(inq_data) - 1]
        )
    )
    # 卡尔曼滤波
    initial_state_mean = 0
    initial_state_covariance = 0.5
    observation_covariance = 0.5
    transition_covariance = 0.5
    kf = KalmanFilter(
        initial_state_mean=initial_state_mean,
        initial_state_covariance=initial_state_covariance,
        observation_covariance=observation_covariance,
        transition_covariance=transition_covariance,
        n_dim_obs=1,
    )
    kf = kf.em(inq_data, n_iter=10)
    inq_Kalman, _ = kf.filter(inq_data)
    inq_Kalman = np.ravel(inq_Kalman)
    result_df = pd.DataFrame(
        {
            "TM": inq_data_df["TM"],
            "INQ_moving": inq_moving_average,
            "INQ_QF": QF,
            "INQ_Kalman": inq_Kalman.flatten(),
            "INQ_orig": inq_data_df["INQ"],
        }
    )
    # 计算误差
    W_bias_rela_moving_list = []
    Q_bias_rela_moving_list = []
    time_bias_moving_list = []
    esm_bias_moving_list = []
    W_bias_rela_Kalman_list = []
    Q_bias_rela_Kalman_list = []
    time_bias_Kalman_list = []
    esm_bias_Kalman_list = []
    W_bias_rela_QF_list = []
    Q_bias_rela_QF_list = []
    time_bias_QF_list = []
    esm_bias_QF_list = []
    inq_moving_filtered = result_df["INQ_moving"]
    inq_QF_filtered = result_df["INQ_QF"]
    inq_Kalman_filtered = result_df["INQ_Kalman"]
    inq_orig_filtered = result_df["INQ_orig"]
    # 滑动平均的误差
    W_bias_rela_moving = (
        abs(inq_moving_filtered.sum() - inq_orig_filtered.sum())
        / inq_orig_filtered.sum()
    )
    Q_bias_rela_moving = (
        abs(inq_moving_filtered.max() - inq_orig_filtered.max())
        / inq_orig_filtered.max()
    )
    time_bias_moving = (
        abs(inq_moving_filtered.argmax() - inq_orig_filtered.argmax())
        / inq_orig_filtered.argmax()
    )
    esm_bias_moving = calculate_esm(inq_moving_filtered, inq_orig_filtered)
    W_bias_rela_moving_list.append(W_bias_rela_moving)
    Q_bias_rela_moving_list.append(Q_bias_rela_moving)
    time_bias_moving_list.append(time_bias_moving)
    esm_bias_moving_list.append(esm_bias_moving)
    # 五点三次的误差
    W_bias_rela_QF = (
        abs(inq_QF_filtered.sum() - inq_orig_filtered.sum()) / inq_orig_filtered.sum()
    )
    Q_bias_rela_QF = (
        abs(inq_QF_filtered.max() - inq_orig_filtered.max()) / inq_orig_filtered.max()
    )
    time_bias_QF = (
        abs(inq_QF_filtered.argmax() - inq_orig_filtered.argmax())
        / inq_orig_filtered.argmax()
    )
    esm_bias_QF = calculate_esm(inq_QF_filtered, inq_orig_filtered)
    W_bias_rela_QF_list.append(W_bias_rela_QF)
    Q_bias_rela_QF_list.append(Q_bias_rela_QF)
    time_bias_QF_list.append(time_bias_QF)
    esm_bias_QF_list.append(esm_bias_QF)
    # 卡尔曼滤波的误差
    W_bias_rela_Kalman = (
        abs(inq_Kalman_filtered.sum() - inq_orig_filtered.sum())
        / inq_orig_filtered.sum()
    )
    Q_bias_rela_Kalman = (
        abs(inq_Kalman_filtered.max() - inq_orig_filtered.max())
        / inq_orig_filtered.max()
    )
    time_bias_Kalman = (
        abs(inq_Kalman_filtered.argmax() - inq_orig_filtered.argmax())
        / inq_orig_filtered.argmax()
    )
    esm_bias_Kalman = calculate_esm(inq_Kalman_filtered, inq_orig_filtered)
    W_bias_rela_Kalman_list.append(W_bias_rela_Kalman)
    Q_bias_rela_Kalman_list.append(Q_bias_rela_Kalman)
    time_bias_Kalman_list.append(time_bias_Kalman)
    esm_bias_Kalman_list.append(esm_bias_Kalman)
    inq_bias_df = pd.DataFrame(
        {
            "TM": inq_data_df["TM"],
            "INQ_moving": inq_moving_average,
            "INQ_QF": QF,
            "INQ_Kalman": inq_Kalman.flatten(),
            "INQ_orig": inq_data_df["INQ"],
        }
    )
    bias = pd.DataFrame(
        {
            # 'TM': inq_data_df['TM'],
            "W_bias_rela_moving_list": W_bias_rela_moving_list,
            "Q_bias_rela_moving_list": Q_bias_rela_moving_list,
            "time_bias_moving_list": time_bias_moving_list,
            "esm_bias_moving_list": esm_bias_moving_list,
            "W_bias_rela_Kalman_list": W_bias_rela_Kalman_list,
            "Q_bias_rela_Kalman_list": Q_bias_rela_Kalman_list,
            "time_bias_Kalman_list": time_bias_Kalman_list,
            "esm_bias_Kalman_list": esm_bias_Kalman_list,
            "W_bias_rela_QF_list": W_bias_rela_QF_list,
            "W_bias_rela_QF_list": W_bias_rela_QF_list,
            "Q_bias_rela_QF_list": Q_bias_rela_QF_list,
            "time_bias_QF_list": time_bias_QF_list,
            "esm_bias_QF_list": esm_bias_QF_list,
        }
    )
    return inq_bias_df



import os

import geopandas as gpd
import h5py
import numpy as np
import pandas as pd
import xarray as xr
from pandas import DataFrame
from pandas.core.indexes.api import default_index
import hydrodatasource.processor.mask as hpm
import hydrodatasource.configs.config as hdscc
from hydrodatasource.processor.basin_mean_rainfall import rainfall_average
from hydrodatasource.reader import access_fs


def query_path_from_metadata(
    basin_id, time_start=None, time_end=None, bbox=None, data_source="gpm"
):
    # query path from other columns from metadata.csv
    metadata_df = pd.read_csv(
        f"s3://grids-origin/{data_source}_metadata.csv",
        storage_options=hdscc.MINIO_PARAM,
    )
    source_list = [
        x
        for x in metadata_df["path"]
        if ((data_source in x) | (data_source.upper() in x))
    ]
    paths = metadata_df[metadata_df["path"].isin(source_list)]
    if time_start is not None:
        paths = paths[paths["time_start"] >= time_start]
    if time_end is not None:
        paths = paths[paths["time_end"] <= time_end]
    if (
        (data_source == "gpm")
        or (data_source == "smap")
        or (data_source == "era5_land")
        or (data_source == "era5")
    ):
        if bbox is not None:
            paths = paths[
                (paths["bbox"].apply(lambda x: string_to_list(x)[0] <= bbox[0]))
                & (paths["bbox"].apply(lambda x: string_to_list(x)[1] >= bbox[1]))
                & (paths["bbox"].apply(lambda x: string_to_list(x)[2] >= bbox[2]))
                & (paths["bbox"].apply(lambda x: string_to_list(x)[3] <= bbox[3]))
            ]
    elif data_source == "gfs":
        path_list_predicate = paths[
            paths["path"].isin(choose_gfs(paths, time_start, time_end))
        ]
        paths = paths[
            path_list_predicate["bbox"].apply(lambda x: string_to_list(x)[0] <= bbox[0])
            & (paths["bbox"].apply(lambda x: string_to_list(x)[1] >= bbox[1]))
            & (paths["bbox"].apply(lambda x: string_to_list(x)[2] >= bbox[2]))
            & (paths["bbox"].apply(lambda x: string_to_list(x)[3] <= bbox[3]))
        ]
    candidate_tile_list = paths[
        paths["path"].apply(lambda x: ("_tile" in x) & (basin_id in x))
    ]
    if len(candidate_tile_list) == 0:
        tile_list = generate_metadata(
            paths, data_source, bbox, time_start, time_end, basin_id
        )[0]
    else:
        tile_list = candidate_tile_list["path"][
            candidate_tile_list["bbox"].apply(lambda x: str(bbox) == x)
        ].to_list()
        should_length = standard_length(data_source, time_start, time_end)
        if len(tile_list) < should_length - 2:
            tile_list = generate_metadata(
                paths, data_source, bbox, time_start, time_end, basin_id
            )[0]
    return tile_list


def generate_metadata(
    paths: DataFrame, data_source: str, bbox: list, time_start, time_end, basin_id
):
    metadata_df = pd.read_csv(
        f"s3://grids-origin/{data_source}_metadata.csv",
        storage_options=hdscc.MINIO_PARAM,
    )
    tile_list = generate_tile_list(
        paths, data_source, bbox, time_start, time_end, basin_id
    )
    bbox_array = np.repeat(str(bbox), len(tile_list))
    time_start_array = np.repeat(str(time_start), len(tile_list))
    time_end_array = np.repeat(str(time_end), len(tile_list))
    if data_source != "smap":
        res_array = np.repeat(hpm.get_para(data_source)[0], len(tile_list))
        temp_df = pd.DataFrame(
            {
                "bbox": bbox_array,
                "time_start": time_start_array,
                "time_end": time_end_array,
                "res_lon": res_array,
                "res_lat": res_array,
                "path": tile_list,
            },
            index=default_index(len(tile_list)),
        )
    else:
        temp_df = pd.DataFrame(
            {
                "bbox": bbox_array,
                "time_start": time_start_array,
                "time_end": time_end_array,
                "res_lon": np.repeat(0.08, len(tile_list)),
                "res_lat": np.repeat(0.08, len(tile_list)),
                "path": tile_list,
            },
            index=default_index(len(tile_list)),
        )
    metadata_df = pd.concat([metadata_df, temp_df], axis=0)
    metadata_df.to_csv(f"{data_source}_metadata.csv", index=False)
    hdscc.FS.put_file(
        f"{data_source}_metadata.csv", f"s3://grids-origin/{data_source}_metadata.csv"
    )
    os.remove(f"{data_source}_metadata.csv")
    return tile_list, metadata_df


def generate_tile_list(
    paths: DataFrame, data_source, bbox, time_start, time_end, basin_id
):
    tile_list = []
    for path in paths["path"]:
        tile_path = path.rstrip(".nc4") + f"{basin_id}_tile.nc4"
        if data_source == "smap":
            path_ds = h5py.File(hdscc.FS.open(path))
            # datetime.fromisoformat('2000-01-01T12:00:00') + timedelta(seconds=path_ds['time'][0])
            lon_array = path_ds["cell_lon"][0]
            lat_array = path_ds["cell_lat"][:, 0]
            cell_lon_w = np.argwhere(lon_array >= bbox[0])[0][0]
            cell_lon_e = np.argwhere(lon_array <= bbox[1])[-1][0]
            cell_lat_n = np.argwhere(lat_array <= bbox[2])[0][0]
            cell_lat_s = np.argwhere(lat_array >= bbox[3])[-1][0]
            tile_da = path_ds["Geophysical_Data"]["sm_surface"][
                cell_lat_n : cell_lat_s + 1, cell_lon_w : cell_lon_e + 1
            ]
            tile_ds = xr.DataArray(tile_da).to_dataset(name="sm_surface")
        elif (data_source == "era5_land") | (data_source == "era5"):
            path_ds = access_fs.spec_path(path.lstrip("s3://"), head="minio")
            tile_ds = path_ds.sel(
                time=slice(time_start, time_end),
                longitude=slice(bbox[0], bbox[1]),
                latitude=slice(bbox[2], bbox[3]),
            )
        else:
            # 会扰乱桶，注意
            path_ds = xr.open_dataset(hdscc.FS.open(path))
            if data_source == "gpm":
                tile_ds = path_ds.sel(
                    time=slice(time_start, time_end),
                    lon=slice(bbox[0], bbox[1]),
                    lat=slice(bbox[3], bbox[2]),
                )
            # 会扰乱桶，注意
            elif data_source == "gfs":
                tile_ds = path_ds.sel(
                    time=slice(time_start, time_end),
                    longitude=slice(bbox[0], bbox[1]),
                    latitude=slice(bbox[2], bbox[3]),
                )
            else:
                tile_ds = path_ds
        if data_source in ["gpm", "smap", "era5_land", "era5"]:
            hdscc.FS.write_bytes(tile_path, tile_ds.to_netcdf())
        elif data_source == "gfs":
            tile_ds.to_netcdf("temp.nc4")
            hdscc.FS.put_file("temp.nc4", tile_path)
            os.remove("temp.nc4")
        tile_list.append(tile_path)
    return tile_list


def standard_length(data_source, time_start, time_end):
    if data_source == "gpm":
        return len(pd.date_range(time_start, time_end, freq="30min").to_list())
    elif data_source == "smap":
        return len(pd.date_range(time_start, time_end, freq="3h").to_list())
    elif (
        (data_source == "gfs") | (data_source == "era5_land") | (data_source == "era5")
    ):
        return len(pd.date_range(time_start, time_end, freq="h").to_list())
    else:
        return len(pd.date_range(time_start, time_end, freq="d").to_list())


def choose_gfs(paths, start_time, end_time):
    """
    This function chooses GFS data within a specified time range and bounding box.
    Args:
        paths (Dataframe): A list of GFS data paths.
        start_time (datetime, YY-mm-dd): The start time of the desired data.
        end_time (datetime, YY-mm-dd): The end time of the desired data.
    Returns:
        list: A list of GFS data within the specified time range and bounding box.
    """
    path_list = []
    produce_times = ["00", "06", "12", "18"]
    if start_time is None:
        start_time = paths["time_start"].iloc[0]
    if end_time is None:
        end_time = paths["time_end"].iloc[-1]
    time_range = pd.date_range(start_time, end_time, freq="1D")
    for date in time_range:
        date_str = date.strftime("%Y/%m/%d")
        for i in range(len(produce_times)):
            for j in range(6 * i, 6 * (i + 1)):
                path = (
                    "s3://grids-origin/GFS/GEE/1h/"
                    + date_str
                    + "/"
                    + produce_times[i]
                    + "/gfs20220103.t"
                    + produce_times[i]
                    + "z.nc4.0p25.f"
                    + "{:03d}".format(j)
                )
                path_list.append(path)
    return path_list


def string_to_list(x: str):
    return list(map(float, x[1:-1].split(",")))


def generate_bbox_from_shp(basin_shape_path, data_source, minio=True):
    # 只考虑单个流域
    if minio:
        basin_gpd = access_fs.spec_path(basin_shape_path.lstrip("s3://"), head="minio")
    else:
        basin_gpd = gpd.read_file(basin_shape_path)
    if data_source == "smap":
        # 西，东，北，南
        bounds_array = basin_gpd.bounds.to_numpy()[0]
        bbox = [bounds_array[0], bounds_array[2], bounds_array[3], bounds_array[1]]
    else:
        mask = hpm.gen_single_mask(
            basin_shape_path.split("/")[-1], basin_gpd, dataname=data_source
        )
        bbox = [
            mask["lon"].values.min(),
            mask["lon"].values.max(),
            mask["lat"].values.max(),
            mask["lat"].values.min(),
        ]
    return bbox, basin_gpd


def grid_mean_mask(basin_id, times: list, data_source):
    # basin_id: basin_CHN_songliao_21401550, 碧流河
    # times: [[2023-06-06 00:00:00, 2023-06-06 02:00:00], [2023-06-07 00:00:00, 2023-06-07 02:00:00]]
    basin_shp = f"s3://basins-origin/basin_shapefiles/{basin_id}.zip"
    bbox, basin = generate_bbox_from_shp(basin_shp, data_source=data_source)
    aver_npy = f"{basin_id}_{times}_{data_source}_hour_array.npy"
    s3_aver_npy_path = f"s3://basins-origin/hour_data/1h/mean_data/{aver_npy}"
    if hdscc.FS.exists(s3_aver_npy_path):
        result_arr_list = np.load(hdscc.FS.open(s3_aver_npy_path), allow_pickle=True)
    else:
        aoi_data_paths = []
        for time_slice in times:
            time_start = time_slice[0]
            time_end = time_slice[1]
            aoi_path = query_path_from_metadata(
                basin_id, time_start, time_end, bbox, data_source=data_source
            )
            aoi_data_paths.append(aoi_path)
        result_arr_list = []
        for time_paths in aoi_data_paths:
            for path in time_paths:
                aoi_dataset = xr.open_dataset(hdscc.FS.open(path))
                mask = hpm.gen_single_mask(basin_id, basin, data_source)
                if data_source == "gpm":
                    # 按照mask出来的四至和get_para()有关，全是.0或.5，直接输入四至就会破坏这样的性质
                    result_arr = hpm.mean_by_mask(
                        aoi_dataset, var="precipitationCal", mask=mask
                    )
                elif (data_source == "era5_land") | (data_source == "era5"):
                    result_arr = hpm.mean_by_mask(aoi_dataset, var="tp", mask=mask)
                elif data_source == "smap":
                    result_arr = hpm.mean_by_mask(
                        aoi_dataset, var="sm_surface", mask=mask
                    )
                elif data_source == "gfs":
                    result_arr = hpm.mean_by_mask(aoi_dataset, var="APCP", mask=mask)
                else:
                    result_arr = []
                result_arr_list.append(result_arr)
        result_arr_list = np.array(result_arr_list)
        np.save(aver_npy, result_arr_list)
        hdscc.FS.put_file(aver_npy, s3_aver_npy_path)
        os.remove(aver_npy)
    return result_arr_list, basin


def concat_gpm_average(basin_id, times: list):
    gpm_aver_npy = f"{basin_id}_{times}_gpm_hour_array_concat.npy"
    s3_gpm_aver_path = "s3://basins-origin/hour_data/1h/mean_data/" + gpm_aver_npy
    if hdscc.FS.exists(s3_gpm_aver_path):
        gpm_hour_array = np.load(hdscc.FS.open(s3_gpm_aver_path), allow_pickle=True)
    else:
        gpm_half_hour_array, basin = grid_mean_mask(basin_id, times, "gpm")
        gpm_hour_array = []
        for i in np.arange(0, len(gpm_half_hour_array) - 1, 2):
            gpm_hour_i = np.add(gpm_half_hour_array[i], gpm_half_hour_array[i + 1])
            gpm_hour_array.append(gpm_hour_i)
        # temporarily fix
        if len(gpm_hour_array) % 2 != 0:
            gpm_hour_array.extend(gpm_hour_array[-1])
        gpm_hour_array = np.array(gpm_hour_array, dtype=object)
        np.save(gpm_aver_npy, gpm_hour_array)
        hdscc.FS.put_file(gpm_aver_npy, s3_gpm_aver_path)
        os.remove(gpm_aver_npy)
    return gpm_hour_array


def concat_gpm_smap_mean_data(basin_ids: list, times: list, use_pp=False):
    pp_sta_gdf = gpd.read_file(
        hdscc.FS.open("s3://stations-origin/stations_list/pp_stations.zip")
    )
    merge_list = []
    for basin_id in basin_ids:
        gpm_mean = concat_gpm_average(basin_id, times)
        smap_mean_mask, basin_gdf = grid_mean_mask(basin_id, times, "smap")
        smap_mean = np.repeat(smap_mean_mask, 3, axis=1).flatten()
        if smap_mean.shape[0] < gpm_mean.shape[0]:
            diff = gpm_mean.shape[0] - smap_mean.shape[0]
            smap_mean = np.pad(smap_mean, (0, diff), "edge")
        else:
            smap_mean = smap_mean[: gpm_mean.shape[0]]
        streamflow_arr = (read_streamflow_from_minio(times, basin_id.lstrip("basin_")))[
            "Q"
        ].to_numpy()
        if streamflow_arr.shape[0] < gpm_mean.shape[0]:
            diff = gpm_mean.shape[0] - streamflow_arr.shape[0]
            streamflow_arr = np.pad(streamflow_arr, (0, diff), "edge")
        if use_pp:
            pp_stas_basin = gpd.overlay(pp_sta_gdf, basin_gdf, "intersection")
            average_rainfall = rainfall_average(
                basin_gdf, pp_stas_basin, pp_stas_basin["ID"].to_list(), times[0][0]
            )
            average_rainfall_times = average_rainfall[
                average_rainfall["TM"].isin(convert_time_slice_to_range(times))
            ]
            temp_df = pd.DataFrame(
                {
                    "time": convert_time_slice_to_range(times),
                    "gpm_tp(mm/h)": gpm_mean,
                    "smap(m3/m3)": smap_mean,
                    "sta_tp(mm/h)": average_rainfall_times[
                        "weighted_rainfall"
                    ].to_numpy(),
                    "streamflow(m3/s)": streamflow_arr,
                }
            ).set_index(keys=["time"])
        else:
            temp_df = pd.DataFrame(
                {
                    "time": convert_time_slice_to_range(times),
                    "gpm_tp(mm/h)": gpm_mean,
                    "smap(m3/m3)": smap_mean,
                    "streamflow(m3/s)": streamflow_arr,
                }
            ).set_index(keys=["time"])
        merge_ds = xr.Dataset.from_dataframe(temp_df.astype("float64"))
        merge_list.append(merge_ds)
        hdscc.FS.write_bytes(
            f"s3://basins-origin/hour_data/1h/mean_data/mean_data_merged/mean_data_{basin_id}.nc",
            merge_ds.to_netcdf(),
        )
    return merge_list


def convert_time_slice_to_range(time_slice_list):
    time_range_list = []
    for time_slice in time_slice_list:
        pd_time_slice = pd.date_range(time_slice[0], time_slice[1], freq="H").to_list()
        time_range_list.extend(pd_time_slice)
    return time_range_list


def read_streamflow_from_minio(times: list, sta_id=""):
    # sta_id: CHN_songliao_21401550、USA_xxx_01301500
    stream_times_df = pd.DataFrame()
    for time_slice in times:
        if (pd.to_datetime("2020-01-01") > pd.to_datetime(time_slice[1])) & (
            "camels" in sta_id
        ):
            stcd = sta_id.split("_")[-1]
            zq_1h_path = f"datasets-origin/camels-hourly/data/usgs_streamflow_csv/{stcd}-usgs-hourly.csv"
            if hdscc.FS.exists(zq_1h_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zq_1h_path), index_col=None, parse_dates=["date"]
                )
                streamflow_df = streamflow_df[
                    streamflow_df["date"].isin(
                        pd.date_range(time_slice[0], time_slice[1], freq="H")
                    )
                ]
                streamflow_df = streamflow_df.rename(
                    columns={"date": "TM", "QObs(mm/h)": "Q"}
                )
                streamflow_df = streamflow_df[["TM", "Q"]]
            else:
                streamflow_df = pd.DataFrame()
        else:
            if "camels" in sta_id:
                stcd = sta_id.split("_")[-1]
                sta_id = f"USA_usgs_{stcd}"
            zq_1h_path = (
                f"s3://stations-origin/zq_stations/hour_data/1h/zq_{sta_id}.csv"
            )
            zq_6h_path = (
                f"s3://stations-origin/zq_stations/hour_data/6h/zq_{sta_id}.csv"
            )
            zq_1d_path = f"s3://stations-origin/zq_stations/day_data/1d/zq_{sta_id}.csv"
            zz_1h_path = (
                f"s3://stations-origin/zz_stations/hour_data/1h/zz_{sta_id}.csv"
            )
            zz_6h_path = (
                f"s3://stations-origin/zz_stations/hour_data/6h/zz_{sta_id}.csv"
            )
            zz_1d_path = f"s3://stations-origin/zz_stations/day_data/1d/zz_{sta_id}.csv"
            if hdscc.FS.exists(zq_1h_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zq_1h_path), index_col=None, parse_dates=["TM"]
                )
            elif hdscc.FS.exists(zq_6h_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zq_6h_path), index_col=None, parse_dates=["TM"]
                )
            elif hdscc.FS.exists(zz_1h_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zz_1h_path), index_col=None, parse_dates=["TM"]
                )
            elif hdscc.FS.exists(zz_6h_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zz_6h_path), index_col=None, parse_dates=["TM"]
                )
            elif hdscc.FS.exists(zz_1d_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zz_1d_path), index_col=None, parse_dates=["TM"]
                )
            elif hdscc.FS.exists(zq_1d_path):
                streamflow_df = pd.read_csv(
                    hdscc.FS.open(zq_1d_path), index_col=None, parse_dates=["TM"]
                )
            else:
                streamflow_df = pd.DataFrame()
            streamflow_df = streamflow_df[
                streamflow_df["TM"].isin(
                    pd.date_range(time_slice[0], time_slice[1], freq="H")
                )
            ]
            streamflow_df = streamflow_df[["TM", "Q"]]
        stream_times_df = pd.concat([stream_times_df, streamflow_df])
    return stream_times_df


"""
def merge_with_spatial_average(gpm_file, gfs_file, smap_file, output_file_path):
    def calculate_and_rename(input_file_path, prefix):
        ds = access_fs.spec_path(input_file_path, head="minio")
        avg_ds = ds.mean(dim=["lat", "lon"], skipna=True).astype("float32")
        new_names = {var_name: (
            f"{prefix}_tp" if var_name in ["tp", "__xarray_dataarray_variable__"] else f"{prefix}_{var_name}") for
            var_name in avg_ds.data_vars}
        avg_ds_renamed = avg_ds.rename(new_names)
        return avg_ds_renamed
    basin_id = output_file_path.split('_')[-1].split('.')[0]
    gfs_avg_renamed = calculate_and_rename(gfs_file, "gfs")
    gpm_avg_renamed = calculate_and_rename(gpm_file, "gpm")
    smap_avg_renamed = calculate_and_rename(smap_file, "smap")
    intersect_time = np.intersect1d(gfs_avg_renamed.time.values, gpm_avg_renamed.time.values, assume_unique=True)
    intersect_time = np.intersect1d(intersect_time, smap_avg_renamed.time.values, assume_unique=True)
    gfs_intersected = gfs_avg_renamed.sel(time=intersect_time)
    gpm_intersected = gpm_avg_renamed.sel(time=intersect_time)
    smap_intersected = smap_avg_renamed.sel(time=intersect_time)
    merged_ds = xr.merge([gfs_intersected, gpm_intersected, smap_intersected])
    merged_ds = merged_ds.assign_coords({"basin": basin_id}).expand_dims("basin")
    conf.FS.write_bytes(output_file_path, merged_ds.to_netcdf())
    return merged_ds
"""



import re
import numpy as np
import geopandas as gpd
from netCDF4 import Dataset, date2num, num2date
import time
from datetime import datetime, timedelta
import pandas as pd
import pint
import xarray as xr
import contextlib
import tempfile
from ..configs.config import FS

# please don't remove the following line although it seems not used
import pint_xarray  # noqa

from hydroutils.hydro_time import calculate_utc_offset


def creatspinc(value, data_vars, lats, lons, starttime, filename, resolution):
    gridspi = Dataset(filename, "w", format="NETCDF4")

    # dimensions
    gridspi.createDimension("time", value[0].shape[0])
    gridspi.createDimension("lat", value[0].shape[2])  # len(lat)
    gridspi.createDimension("lon", value[0].shape[1])

    # Create coordinate variables for dimensions
    times = gridspi.createVariable("time", np.float64, ("time",))
    latitudes = gridspi.createVariable("lat", np.float32, ("lat",))
    longitudes = gridspi.createVariable("lon", np.float32, ("lon",))

    # Create the actual variable
    for var, attr in data_vars.items():
        gridspi.createVariable(
            var,
            np.float32,
            (
                "time",
                "lon",
                "lat",
            ),
        )

    # Global Attributes
    gridspi.description = "var"
    gridspi.history = f"Created {time.ctime(time.time())}"
    gridspi.source = "netCDF4 python module tutorial"

    # Variable Attributes
    latitudes.units = "degree_north"
    longitudes.units = "degree_east"
    times.units = "days since 1970-01-01 00:00:00"
    times.calendar = "gregorian"

    # data
    latitudes[:] = lats
    longitudes[:] = lons

    # Fill in times
    dates = []
    if resolution == "daily":
        for n in range(value[0].shape[0]):
            dates.append(starttime + n)
        times[:] = dates[:]

    elif resolution == "6-hourly":
        # for n in range(value[0].shape[0]):
        #     dates.append(starttime + (n+1) * np.timedelta64(6, 'h'))

        for n in range(value[0].shape[0]):
            dates.append(starttime + (n + 1) * timedelta(hours=6))

        times[:] = date2num(dates, units=times.units, calendar=times.calendar)
        # print 'time values (in units %s): ' % times.units +'\n', times[:]
        dates = num2date(times[:], units=times.units, calendar=times.calendar)

    # Fill in values
    i = 0
    for var, attr in data_vars.items():
        gridspi.variables[var].long_name = attr["long_name"]
        gridspi.variables[var].units = attr["units"]
        gridspi.variables[var][:] = value[i][:]
        i = i + 1

    gridspi.close()


def regen_box(bbox, resolution, offset):
    lx = bbox[0]
    rx = bbox[2]
    LLON = np.round(
        int(lx)
        + resolution * int((lx - int(lx)) / resolution + 0.5)
        + offset
        * (int(lx * 10) / 10 + offset - lx)
        / abs(int(lx * 10) // 10 + offset - lx + 0.0000001),
        3,
    )
    RLON = np.round(
        int(rx)
        + resolution * int((rx - int(rx)) / resolution + 0.5)
        - offset
        * (int(rx * 10) / 10 + offset - rx)
        / abs(int(rx * 10) // 10 + offset - rx + 0.0000001),
        3,
    )

    by = bbox[1]
    ty = bbox[3]
    BLAT = np.round(
        int(by)
        + resolution * int((by - int(by)) / resolution + 0.5)
        + offset
        * (int(by * 10) / 10 + offset - by)
        / abs(int(by * 10) // 10 + offset - by + 0.0000001),
        3,
    )
    TLAT = np.round(
        int(ty)
        + resolution * int((ty - int(ty)) / resolution + 0.5)
        - offset
        * (int(ty * 10) / 10 + offset - ty)
        / abs(int(ty * 10) // 10 + offset - ty + 0.0000001),
        3,
    )

    # print(LLON,BLAT,RLON,TLAT)
    return [LLON, BLAT, RLON, TLAT]


def validate(date_text, formatter, error):
    try:
        return datetime.strptime(date_text, formatter)
    except ValueError as e:
        raise ValueError(error) from e


def cf2datetime(ds):
    ds = ds.copy()
    time_tmp1 = ds.indexes["time"]
    attrs = ds.coords["time"].attrs
    time_tmp2 = []
    for i in range(time_tmp1.shape[0]):
        tmp = time_tmp1[i]
        a = str(tmp.year).zfill(4)
        b = str(tmp.month).zfill(2)
        c = str(tmp.day).zfill(2)
        d = str(tmp.hour).zfill(2)
        e = str(tmp.minute).zfill(2)
        f = str(tmp.second).zfill(2)
        time_tmp2.append(np.datetime64(f"{a}-{b}-{c} {d}:{e}:{f}.00000000"))
    ds = ds.assign_coords(time=time_tmp2)
    ds.coords["time"].attrs = attrs

    return ds


def generate_time_intervals(start_date, end_date):
    # Initialize an empty list to store the intervals
    intervals = []

    # Loop over days
    while start_date <= end_date:
        # Loop over the four time intervals in a day
        intervals.extend(
            [start_date.strftime("%Y-%m-%d"), hour] for hour in ["00", "06", "12", "18"]
        )
        # Move to the next day
        start_date += timedelta(days=1)

    return intervals


def _convert_target_unit(target_unit):
    """Convert user-friendly unit to standard unit for internal calculations."""
    if match := re.match(r"mm/(\d+)(h|d)", target_unit):
        num, unit = match.groups()
        return int(num), unit
    return None, None


def streamflow_unit_conv(streamflow, area, target_unit="mm/d", inverse=False):
    """Convert the unit of streamflow data to mm/xx(time) for a basin or inverse.

    Parameters
    ----------
    streamflow: xarray.Dataset, numpy.ndarray, pandas.DataFrame/Series
        Streamflow data of each basin.
    area: xarray.Dataset or pint.Quantity wrapping numpy.ndarray, pandas.DataFrame/Series
        Area of each basin.
    target_unit: str
        The unit to convert to.
    inverse: bool
        If True, convert the unit to m^3/s.
        If False, convert the unit to mm/day or mm/h.

    Returns
    -------
    Converted data in the same type as the input streamflow.
    """
    # Convert the user input unit format
    num, unit = _convert_target_unit(target_unit)
    if unit:
        if unit == "h":
            standard_unit = "mm/h"
            conversion_factor = num
        elif unit == "d":
            standard_unit = "mm/d"
            conversion_factor = num
        else:
            raise ValueError(f"Unsupported unit: {unit}")
    else:
        standard_unit = target_unit
        conversion_factor = 1
    # Regular expression to match units with numbers
    custom_unit_pattern = re.compile(r"mm/(\d+)(h|d)")

    # Function to handle the conversion for numpy and pandas
    def np_pd_conversion(streamflow, area, target_unit, inverse, conversion_factor):
        if not inverse:
            result = (streamflow / area).to(target_unit) * conversion_factor
        else:
            result = (streamflow * area).to(target_unit) / conversion_factor
        return result.magnitude

    # Handle xarray
    if isinstance(streamflow, xr.Dataset) and isinstance(area, xr.Dataset):
        streamflow_units = streamflow[list(streamflow.keys())[0]].attrs.get(
            "units", None
        )
        if not inverse:
            if not (
                custom_unit_pattern.match(target_unit)
                or re.match(r"mm/(?!\d)", target_unit)
            ):
                raise ValueError(
                    "target_unit should be a valid unit like 'mm/d', 'mm/day', 'mm/h', 'mm/hour', 'mm/3h', 'mm/5d'"
                )

            q = streamflow.pint.quantify()
            a = area.pint.quantify()
            r = q[list(q.keys())[0]] / a[list(a.keys())[0]]
            # result = r.pint.to(target_unit).to_dataset(name=list(q.keys())[0])
            result = (r.pint.to(standard_unit) * conversion_factor).to_dataset(
                name=list(q.keys())[0]
            )
            # Manually set the unit attribute to the custom unit
            result_ = result.pint.dequantify()
            result_[list(result_.keys())[0]].attrs["units"] = target_unit
            return result_
        else:
            if streamflow_units:
                if custom_match := custom_unit_pattern.match(streamflow_units):
                    num, unit = custom_match.groups()
                    if unit == "h":
                        standard_unit = "mm/h"
                        conversion_factor = int(num)
                    elif unit == "d":
                        standard_unit = "mm/d"
                        conversion_factor = int(num)
                    # Convert custom unit to standard unit
                    r_ = streamflow / conversion_factor
                    r_[list(r_.keys())[0]].attrs["units"] = standard_unit
                    r = r_.pint.quantify()
                else:
                    r = streamflow.pint.quantify()
            else:
                r = streamflow.pint.quantify()
            if target_unit not in ["m^3/s", "m3/s"]:
                raise ValueError("target_unit should be 'm^3/s'")
            a = area.pint.quantify()
            q = r[list(r.keys())[0]] * a[list(a.keys())[0]]
            result = q.pint.to(target_unit).to_dataset(name=list(r.keys())[0])
            # dequantify to get normal xr_dataset
            return result.pint.dequantify()

    # Handle numpy and pandas
    elif isinstance(streamflow, pint.Quantity) and isinstance(area, pint.Quantity):
        if type(streamflow.magnitude) not in [np.ndarray, pd.DataFrame, pd.Series]:
            raise TypeError(
                "Input streamflow must be xarray.Dataset, or pint.Quantity wrapping numpy.ndarray, or pandas.DataFrame/Series"
            )
        if type(area.magnitude) != type(streamflow.magnitude):
            raise TypeError("streamflow and area must be the same type")
        return np_pd_conversion(
            streamflow, area, standard_unit, inverse, conversion_factor
        )

    else:
        raise TypeError(
            "Input streamflow must be xarray.Dataset, or pint.Quantity wrapping numpy.ndarray, or pandas.DataFrame/Series"
        )


def minio_file_list(minio_folder_url):
    """
    Get all filenames in a specified directory on MinIO.

    Parameters
    ----------
    minio_folder_url : str
        the minio file url, must start with s3://

    Returns
    -------
    folder list
    """
    # Get the list of files in the directory
    try:
        # the minio folder url doesn't have to start with s3://, but we agree that it must
        # start with s3:// to distinguish between local and Minio folder directories.
        files = FS.ls(minio_folder_url)
        return [file.split("/")[-1] for file in files if not file.endswith("/")]
    except Exception as e:
        print(f"Error accessing {minio_folder_url}: {e}")
        return []


def is_minio_folder(minio_url):
    """
    Check if a MinIO folder exists.

    Parameters
    ----------
    minio_url : str
        the minio file url, must start with s3://

    Returns
    -------
    bool
        True if the folder exists, False otherwise

    """
    try:
        if not FS.exists(minio_url):
            raise FileNotFoundError(f"No file or folder found in {minio_url}")
        if minio_url.endswith("/"):
            # If the path ends with '/', treat it as a directory
            return True
        # Try to list objects under this path
        objects = FS.ls(minio_url)
        test_object = "s3://" + objects[0]
        return len(objects) != 1 or test_object != minio_url
    except Exception as e:
        raise NotImplementedError(f"Error accessing {minio_url}: {e}") from e


def calculate_basin_offsets(shp_file_path):
    """
    Calculate the UTC offset for each basin based on the outlet shapefile.

    Parameters:
        shp_file (str): The path to the basin outlet shapefile.

    Returns:
        dict: A dictionary where the keys are the BASIN_ID and the values are the corresponding UTC offsets.
    """
    # read shapefile
    if "s3://" in shp_file_path:
        # related list
        extensions = [".shp", ".shx", ".dbf", ".prj"]

        # create a temporary directory
        with tempfile.TemporaryDirectory() as tmpdir:
            # download all related files to the temporary directory
            base_name = shp_file_path.rsplit(".", 1)[0]
            extensions = [".shp", ".shx", ".dbf", ".prj"]

            for ext in extensions:
                remote_file = f"{base_name}{ext}"
                local_file = f"{tmpdir}/shp_file{ext}"
                with contextlib.suppress(FileNotFoundError):
                    FS.get(remote_file, local_file)
            gdf = gpd.read_file(f"{tmpdir}/shp_file.shp")

    else:
        # If the file is not on S3 (MinIO), read it directly
        gdf = gpd.read_file(shp_file_path)

    # create an empty dictionary
    basin_offset_dict = {}

    for index, row in gdf.iterrows():
        outlet = row["geometry"]
        offset = calculate_utc_offset(outlet.y, outlet.x)
        basin_id = row.get(
            "BASIN_ID", index
        )  # Use the index as the default value if "BASIN_ID" is not found
        basin_offset_dict[basin_id] = offset

    return basin_offset_dict



import geopandas as gpd
import pandas as pd
import numpy as np
import xarray as xr
import hydrodatasource.configs.config as conf
from hydrodatasource.reader import access_fs
from hydrodatasource.cleaner import rain_anomaly


def test_read_spec():
    # access_fs.spec_path("st_rain_c.csv")
    mean_forcing_nc = access_fs.spec_path(
        "basins-origin/hour_data/1h/mean_data/mean_data_forcing/mean_forcing_CHN_21401550.nc",
        head="minio",
    )
    print(mean_forcing_nc)


def test_read_shp():
    watershed = gpd.read_file(
        conf.FS.open(
            "s3://basins-origin/basin_shapefiles/basin_USA_camels_01411300.zip"
        )
    )
    print(watershed)

    all_watershed = gpd.read_file(conf.FS.open("s3://basins-origin/basins_shp.zip"))
    print(all_watershed)


def test_read_BA():
    basin = HydroBasins(data_path="./")  # 该路径只是为了实例化该类，测试时可随意指定
    attr = basin.read_BA_xrdataset(
        gage_id_lst=["21401550"], var_lst=["all"], path="basins-origin/attributes.nc"
    )
    print(attr.compute())

    all_attr = access_fs.spec_path(
        "basins-origin/attributes.nc",
        head="minio",
    )
    print(all_attr.compute())


def test_read_zz_stations_ts():
    # 读取csv文件
    zz_stations = access_fs.spec_path(
        "stations-origin/zz_stations/hour_data/1h/zz_CHN_songliao_10800100.csv",
        head="minio",
    )
    print(zz_stations)


def test_read_stations_shp():
    zz_stations_gdf = _extracted_from_test_read_stations_shp_3(
        "s3://stations-origin/stations_list/zz_stations.zip",
        "zz_stations 站点列表如下:",
    )
    pp_stations_gdf = _extracted_from_test_read_stations_shp_3(
        "s3://stations-origin/stations_list/pp_stations.zip",
        "pp_stations 站点列表如下:",
    )
    zq_stations_gdf = _extracted_from_test_read_stations_shp_3(
        "s3://stations-origin/stations_list/zq_stations.zip",
        "zq_stations 站点列表如下:",
    )
    return zz_stations_gdf, pp_stations_gdf, zq_stations_gdf


# TODO Rename this here and in `test_read_stations_shp`
def _extracted_from_test_read_stations_shp_3(arg0, arg1):
    # 读取zip中的shpfiles文件
    result = gpd.read_file(conf.FS.open(arg0))
    print(arg1)
    print(result)
    return result


def test_read_stations_list():
    zz_stations_df = _extracted_from_test_read_stations_list_3(
        "s3://stations-origin/stations_list/zz_stations.csv",
        "zz_stations 站点列表如下:",
    )
    pp_stations_df = _extracted_from_test_read_stations_list_3(
        "s3://stations-origin/stations_list/pp_stations.csv",
        "pp_stations 站点列表如下:",
    )
    zq_stations_df = _extracted_from_test_read_stations_list_3(
        "s3://stations-origin/stations_list/zq_stations.csv",
        "zq_stations 站点列表如下:",
    )
    return zz_stations_df, pp_stations_df, zq_stations_df


# TODO Rename this here and in `test_read_stations_list`
def _extracted_from_test_read_stations_list_3(arg0, arg1):
    # 读取csv文件
    result = pd.read_csv(arg0, storage_options=conf.MINIO_PARAM, index_col=False)
    print(arg1)
    print(result)
    return result


def test_read_zqstations_ts():
    return pd.read_csv(
        "s3://stations-origin/zq_stations/zq_CHN_songliao_10310500.csv",
        storage_options=conf.MINIO_PARAM,
    )


def test_read_reservoirs_info():
    dams_gdf = gpd.read_file(conf.FS.open("s3://reservoirs-origin/dams.zip"))
    rsvrs_gdf = gpd.read_file(conf.FS.open("s3://reservoirs-origin/rsvrs_shp.zip"))
    return dams_gdf, rsvrs_gdf


def test_read_river_network():
    return gpd.read_file(conf.FS.open("s3://basins-origin/HydroRIVERS_v10_shp.zip"))


def test_read_rsvr_ts():
    return pd.read_csv(
        "s3://reservoirs-origin/rr_stations/zq_CHN_songliao_10310500.csv",
        storage_options=conf.MINIO_PARAM,
    )


def test_read_pp():
    return pd.read_csv(
        "s3://stations-origin/pp_stations/hour_data/1h/pp_CHN_songliao_10951870.csv",
        storage_options=conf.MINIO_PARAM,
    )


def test_read_zz():
    return pd.read_csv(
        "s3://stations-origin/zz_stations/hour_data/1h/zz_CHN_dalianxiaoku_21302120.csv",
        storage_options=conf.MINIO_PARAM,
    )


def test_read_zq():
    return pd.read_csv(
        "s3://stations-origin/zq_stations/hour_data/1h/zq_USA_usgs_01181000.csv",
        storage_options=conf.MINIO_PARAM,
    )


def test_df2ds():
    zq_df = pd.read_csv(
        "s3://stations-origin/zq_stations/hour_data/1h/zq_USA_usgs_01181000.csv",
        storage_options=conf.MINIO_PARAM,
    )
    return xr.Dataset().from_dataframe(zq_df)


def list_csv_files(bucket_name, prefix=""):
    """List paths of all CSV files in the specified S3 bucket."""
    path = f"{bucket_name}/{prefix}" if prefix else bucket_name
    return conf.FS.glob(f"{path}*.csv")


def process_store_csvs(source_bucket, destination_bucket, prefix=""):
    """
    Read CSV files from the source bucket, process them, and store the results in the destination bucket.
    """
    csv_files = list_csv_files(source_bucket, prefix)

    for file_path in csv_files:
        df = read_csv_to_dataframe(file_path)
        df = process_dataframe(df)
        store_dataframe_to_bucket(df, destination_bucket, file_path)


def read_csv_to_df(file_path):
    """Read CSV file from file_path and return as pandas DataFrame."""
    with conf.FS.open(file_path, mode="rb") as csv_file:
        df = pd.read_csv(csv_file, index_col=None)
    return df


def process_dataframe(df):
    df = rain_anomaly.normalize_rainfall_format(df)
    df = rain_anomaly.filter_rainfall_extremes(df)
    df = rain_anomaly.filter_rainfall_gradients(df)
    df = df[
        [
            "station_code",
            "time",
            "daily_rainfall",
            "interval",
            "precipitation",
            "daily_change",
            "wetness",
        ]
    ]
    return df


def store_dataframe_to_bucket(df, bucket, file_path):
    file_name = file_path.split("/")[-1]
    destination_file_path = f"{bucket}/{file_name}"
    with conf.FS.open(destination_file_path, mode="w") as file:
        df.to_csv(file, index=False)


def test_read_folder():
    source_bucket = "s3://stations-origin/pp_stations/hour_data/1h"
    destination_bucket = "s3://stations-interim/pp_stations/hour_data/1h"
    process_store_csvs(source_bucket, destination_bucket)


def test_read_era5_land_csv():
    era5_land_zarr_files = [
        file
        for file in conf.FS.glob("s3://grids-origin/era5_land/")
        if file.endswith(".zarr")
    ]
    bbox_list = []
    start_time = []
    end_list = []
    res_lon_list = []
    res_lat_list = []
    for i in range(len(era5_land_zarr_files)):
        test_ds_i = xr.open_dataset(conf.FS.open(era5_land_zarr_files[i]))
        bbox = [
            np.min(test_ds_i["longitude"].to_numpy()),
            np.max(test_ds_i["longitude"].to_numpy()),
            np.max(test_ds_i["latitude"].to_numpy()),
            np.min(test_ds_i["latitude"].to_numpy()),
        ]
        bbox_list.append(bbox)
        start_time.append(test_ds_i["time"].to_numpy()[0])
        end_list.append(test_ds_i["time"].to_numpy()[-1])
        lon_res = abs(np.diff(test_ds_i["longitude"].to_numpy())[0])
        res_lon_list.append(lon_res)
        lat_res = abs(np.diff(test_ds_i["latitude"].to_numpy())[0])
        res_lat_list.append(lat_res)
    test_pd = pd.DataFrame(
        {
            "bbox": bbox_list,
            "time_start": start_time,
            "time_end": end_list,
            "res_lon": lon_res,
            "res_lat": lat_res,
            "path": ["s3://" + file for file in era5_land_zarr_files],
        }
    )
    test_pd.to_csv("era5_land_metadata.csv")



"""
Author: Jiaxu Wu
Date: 2024-01-12 15:21:20
LastEditTime: 2024-03-28 08:37:54
LastEditors: Wenyu Ouyang
Description: processor for era5 data
FilePath: \hydrodatasource\hydrodatasource\processor\era5.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""
import pandas as pd
import geopandas as gpd
import xarray as xr

from hydro_gistools.mean import gen_mask, mean_by_mask

from hydrodatasource.reader.minio import ERA5LReader

def get_era5():
    era5_land = ERA5LReader()

    data_variables = ["Total precipitation", "2 metre temperature"]
    start_time = pd.to_datetime("2016-01-01")
    end_time = pd.to_datetime("2020-12-31")
    shp = "/ftproot/LSTM_data/碧流河流域.shp"
    save_file = "data/era5_land_5.nc"

    era5_land.to_netcdf(
        data_variables=data_variables,
        start_time=start_time,
        end_time=end_time,
        shp=shp,
        save_file=save_file,
    )


def get_mean():
    watershed = gpd.read_file("/ftproot/LSTM_data/碧流河流域.shp")
    # hydro_gistools mean.py line 249 overlay() keep_geom_type=False without warning
    gen_mask(watershed, "FID", "gfs", save_dir="data/mask")
    gen_mask(watershed, "FID", "gpm", save_dir="data/mask")
    gen_mask(watershed, "FID", "era5", save_dir="data/mask")
    data = xr.open_dataset("/ftproot/LSTM_data/era5_land_5.nc")
    mask = xr.open_dataset("data/mask/mask-1-era5.nc")
    mean_tp = mean_by_mask(data, "tp", mask)
    print("mean_tp computed!")
    mean_t2m = mean_by_mask(data, "t2m", mask)
    print("mean_t2m computed!")
    df = pd.DataFrame()
    df["TM"] = data["time"]
    df["tp"] = mean_tp.compute()
    df["t2m"] = mean_t2m.compute()
    df.to_csv("data/era5_mean.csv", index=False)
    print("data/era5_mean.csv saved!")
    basin_value = "21401550"
    ds = xr.Dataset(
        {
            "prcp": (("time", "basin"), df[["tp"]].values),
            "temperature": (("time", "basin"), df[["t2m"]].values),
        },
        coords={"time": df["TM"].values, "basin": [basin_value]},
    )
    ds["time"] = ds["time"].astype("datetime64[ns]")
    nc_file_path = "data/era5_mean.nc"
    ds.to_netcdf(nc_file_path)
    print("data/era5_mean.nc saved!")


get_mean()



'''
Author: liutiaxqabs 1498093445@qq.com
Date: 2024-05-28 10:24:16
LastEditors: liutiaxqabs 1498093445@qq.com
LastEditTime: 2024-06-22 10:40:23
FilePath: /hydrodatasource/tests/test_rainfall_cleaner.py
Description: 这是默认设置,请设置`customMade`, 打开koroFileHeader查看配置 进行设置: https://github.com/OBKoro1/koro1FileHeader/wiki/%E9%85%8D%E7%BD%AE
'''

import pytest
from hydrodatasource.cleaner.rainfall_cleaner import RainfallCleaner, RainfallAnalyzer

import pandas as pd
import matplotlib.pyplot as plt
import os
from tqdm import tqdm


def test_anomaly_process():
    # 测试降雨数据处理功能
    cleaner = RainfallCleaner(
        data_path="/ftproot/tests_stations_anomaly_detection/rainfall_cleaner/pp_CHN_songliao_21422982.csv",
        era5_path="/ftproot/tests_stations_anomaly_detection/era5land/",
        station_file="/ftproot/tests_stations_anomaly_detection/stations/pp_stations.csv",
        start_time="2020-01-01",
        end_time="2022-10-07",
    )
    # methods默认可以联合调用，也可以单独调用。大多数情况下，默认调用detect_sum
    methods = ["detect_sum"]
    cleaner.anomaly_process(methods)

    print(cleaner.origin_df)
    print(cleaner.processed_df)
    cleaner.processed_df.to_csv(
        "/ftproot/tests_stations_anomaly_detection/results/sampledatatest.csv"
    )
    cleaner.temporal_list.to_csv(
        "/ftproot/tests_stations_anomaly_detection/results/temporal_list.csv"
    )
    cleaner.spatial_list.to_csv(
        "/ftproot/tests_stations_anomaly_detection/results/spatial_list.csv"
    )


def test_basins_polygon_mean():
    # 测试泰森多边形平均值，碧流河为例。测试结果见 /ftproot/tests_stations_anomaly_detection/plot
    basins_mean = RainfallAnalyzer(
        stations_csv_path="/ftproot/basins-origin/basins_pp_stations/20600340_stations.csv",# 站点表，其中ID列带有前缀‘pp_’
        shp_folder="/ftproot/basins-origin/basins_shp/20600340/",
        rainfall_data_folder="/ftproot/basins-origin/basins_songliao_pp_origin_data/20600340/",
        output_folder="/ftproot/basins-origin/basins_rainfall_mean/",
        output_log="/ftproot/basins-origin/basins_rainfall_mean/plot/summary_log.txt",
        output_plot="/ftproot/basins-origin/basins_rainfall_mean/plot/",
        lower_bound=200,
        upper_bound=2000,
    )
    basins_mean.basins_polygon_mean()


def test_basins_polygon_mean_folder():
    # 设置基础路径
    base_shp_folder = "/ftproot/basins-origin/basins_shp/"
    base_stations_csv_folder = "/ftproot/basins-origin/basins_pp_stations/"
    base_rainfall_data_folder = "/ftproot/basins-origin/basins_songliao_pp_origin_data/"
    output_folder = "/ftproot/basins-origin/basins_rainfall_mean/"
    output_log = os.path.join(output_folder, "plot", "summary_log.txt")
    output_plot = os.path.join(output_folder, "plot")

    # 获取shp_folder目录下的所有文件夹名称并排序
    subfolders = sorted([f.name for f in os.scandir(base_shp_folder) if f.is_dir()])

    # 遍历所有文件夹并运行 basins_polygon_mean
    for subfolder in tqdm(subfolders, desc="Processing basins"):
        stations_csv_path = os.path.join(base_stations_csv_folder, f"{subfolder}_stations.csv")
        shp_folder = os.path.join(base_shp_folder, subfolder)
        rainfall_data_folder = os.path.join(base_rainfall_data_folder, subfolder)
        
        if os.path.exists(stations_csv_path) and os.path.exists(shp_folder) and os.path.exists(rainfall_data_folder):
            try:
                basins_mean = RainfallAnalyzer(
                    stations_csv_path=stations_csv_path,
                    shp_folder=shp_folder,
                    rainfall_data_folder=rainfall_data_folder,
                    output_folder=output_folder,
                    output_log=output_log,
                    output_plot=output_plot,
                    lower_bound=200,
                    upper_bound=2000,
                )
                basins_mean.basins_polygon_mean()
            except Exception as e:
                print(f"Error processing {subfolder}: {e}")
                # 这里可以添加更多调试信息，比如打印 DataFrame 的列名
        else:
            print(f"Missing required files for {subfolder}")


def test_time_consistency():
    time_consistency_check = RainfallAnalyzer(
        rainfall_data_folder="/ftproot/tests_stations_anomaly_detection/rainfall_cleaner/",
        output_folder="/ftproot/tests_stations_anomaly_detection/results/",
    )
    time_consistency_check.time_consistency()
    
def test_spatial_consistency():
    basins_spatial = RainfallAnalyzer(
        stations_csv_path="/ftproot/tests_stations_anomaly_detection/stations/pp_stations.csv",
        shp_folder="/ftproot/tests_stations_anomaly_detection/shapefiles/",
        rainfall_data_folder="/ftproot/tests_stations_anomaly_detection/rainfall_cleaner/",
        output_folder="/ftproot/tests_stations_anomaly_detection/results/",
        output_plot="/ftproot/tests_stations_anomaly_detection/plot/",
    )
    basins_spatial.spatial_consistency()


"""Main module."""
import asyncio
import os
import pathlib

from minio import Minio
import chardet


async def minio_upload_csv(client, bucket_name, object_name, file_path):
    """upload csv to minio

    Parameters
    ----------
    client : _type_
        the minio client
    bucket_name : _type_
        the bucket name
    object_name : _type_
        the object name
    file_path : _type_
        the local file path
    """
    # Make a bucket
    bucket_names = [bucket.name for bucket in client.list_buckets()]
    if bucket_name not in bucket_names:
        client.make_bucket(bucket_name)
    # Upload an object
    client.fput_object(bucket_name, object_name, file_path)
    # List objects
    objects = client.list_objects(bucket_name, recursive=True)
    return [obj.object_name for obj in objects]


async def minio_download_csv(client: Minio, bucket_name, object_name: str, file_path: str, version_id=None):
    try:
        response = client.get_object(bucket_name, object_name, version_id)
        # 图片不能直接解码, 但可以直接对文件写入response.data
        encoding = chardet.detect(response.data)['encoding']
        object_path = os.path.join(file_path, object_name)
        object_parent = pathlib.Path(object_path).parent
        if not object_parent.exists():
            object_parent.mkdir(parents=True)
        if encoding is not None:
            res_csv: str = response.data.decode(encoding)
            with open(object_path, 'w+', encoding=encoding) as fp:
                fp.write(res_csv)
        else:
            with open(object_path, 'wb') as fp:
                fp.write(response.data)
    finally:
        response.close()
        response.release_conn()


async def boto3_upload_csv(client, bucket_name, object_name, file_path):
    """upload csv to minio

    Parameters
    ----------
    client : _type_
        the minio client
    bucket_name : _type_
        the bucket name
    object_name : _type_
        the object name
    file_path : _type_
        the local file path
    """
    # Make a bucket
    bucket_names = [dic['Name'] for dic in client.list_buckets()['Buckets']]
    if bucket_name not in bucket_names:
        client.create_bucket(Bucket=bucket_name)
    # Upload an object
    client.upload_file(file_path, bucket_name, object_name)
    # List objects
    objects = [dic['Key'] for dic in client.list_objects(Bucket=bucket_name)['Contents']]
    return objects


async def boto3_download_csv(client, bucket_name, object_name, file_path: str):
    client.download_file(bucket_name, object_name, file_path)


async def boto3_sync_files(client, bucket_name, local_path, bucket_path=None):
    """
    :param client: the boto3 client
    :param bucket_name: the bucket name which you want to sync your data
    :param local_path: the path on your local machine
    :param bucket_path: the path under your bucket which you want to sync
    :return:
    """
    remote_objects = [dic['Key'] for dic in client.list_objects(Bucket=bucket_name)['Contents']]
    local_objects = [str(path.relative_to(local_path)).replace('\\', '/')
                     for path in pathlib.Path(local_path).rglob(pattern='*') if path.is_file()]
    objects_in_remote = [obj for obj in remote_objects if obj not in local_objects]
    objects_in_local = [obj for obj in local_objects if obj not in remote_objects]
    task_batch_dload = asyncio.create_task(boto3_batch_download(objects_in_remote, client, bucket_name, local_path))
    task_batch_upload = asyncio.create_task(boto3_batch_upload(objects_in_local, client, bucket_name, local_path))
    await asyncio.gather(task_batch_upload, task_batch_dload)


async def minio_batch_download(objects_in_remote, client: Minio, bucket_name, local_path):
    # 将下载任务打包成大量task容易造成许多上下文切换进而挤兑，目前看来不如顺序下载
    for obj in objects_in_remote:
        await minio_download_csv(client, bucket_name, obj, local_path)


async def minio_batch_upload(objects_in_local, client: Minio, bucket_name, local_path):
    for obj in objects_in_local:
        await minio_upload_csv(client, bucket_name, obj, local_path)


async def boto3_batch_download(objects_in_remote, client, bucket_name, local_path):
    # 将下载任务打包成大量task容易造成许多上下文切换进而挤兑，目前看来不如顺序下载
    for obj in objects_in_remote:
        await boto3_download_csv(client, bucket_name, obj, local_path)


async def boto3_batch_upload(objects_in_remote, client, bucket_name, local_path):
    for obj in objects_in_remote:
        await boto3_upload_csv(client, bucket_name, obj, local_path)


async def minio_sync_files(client: Minio, bucket_name, local_path, bucket_path=None):
    """
    :param client: the minio client
    :param bucket_name: the bucket name which you want to sync your data
    :param local_path: the path on your local machine, contents of bucket_name will be saved to the directory directly
    :param bucket_path: the path under your bucket which you want to sync
    :return:
    """
    remote_objects = [obj.object_name for obj in client.list_objects(bucket_name, prefix=bucket_path, recursive=True)]
    local_objects = [str(path.relative_to(local_path)).replace('\\', '/')
                     for path in pathlib.Path(local_path).rglob(pattern='*') if path.is_file()]
    objects_in_remote = [obj for obj in remote_objects if obj not in local_objects]
    objects_in_local = [obj for obj in local_objects if obj not in remote_objects]
    task_batch_dload = asyncio.create_task(minio_batch_download(objects_in_remote, client, bucket_name, local_path))
    task_batch_upload = asyncio.create_task(minio_batch_upload(objects_in_local, client, bucket_name, local_path))
    await asyncio.gather(task_batch_upload, task_batch_dload)



#!/usr/bin/env python
# coding: utf-8

"""
该模块用于计算流域平均

- `mean_over_basin` - 计算流域平均

"""

import numpy as np
import xarray as xr
import geopandas as gpd
import pandas as pd
# import dask.array as da
import itertools
from geopandas import GeoDataFrame
from shapely.geometry import Polygon
import hydrodatasource.configs.config as hdscc


def mean_over_basin(basin, basin_id, dataset, data_name, lon="lon", lat="lat"):
    """
    计算流域平均

    Args:
        basin (GeoDataframe): 必选，流域的矢量数据，通过geopandas读取
        basin_id (str): 必选，表示流域编号的字段名称
        dataset (DataArray): 必选，表示grid数据集，通过xarray读取，只含有变量和经纬度
        data_name (str): 必选，表示grid数据集中需要参与计算的变量名称
        lon (str): 可选，grid数据集中经度坐标名称
        lat (str): 可选，grid数据集中纬度坐标名称

    Returns
        data (Dataframe): 流域编号和对应的平均值

    """

    grid = grid_to_gdf(dataset, data_name, lon=lon, lat=lat)
    intersects = gpd.overlay(grid, basin, how="intersection")
    intersects = intersects.to_crs(epsg=3857)
    intersects["Area"] = intersects.area
    intersects = intersects.to_crs(epsg=4326)
    return intersects.groupby(basin_id).apply(wavg, data_name, "Area")


def wavg(group, avg_name, weight_name):
    d = group[avg_name]
    w = group[weight_name]
    try:
        return (d * w).sum() / w.sum()
    except ZeroDivisionError:
        return d.mean()


def grid_to_gdf(dataset, data_name, lon, lat):
    lons = dataset[lon].values
    lats = dataset[lat].values
    delta = lons[1] - lons[0]

    geometry = []
    values = []
    HBlons = []
    HBlats = []

    delta_lon = lons.size
    delta_lat = lats.size

    for i, j in itertools.product(range(delta_lon), range(delta_lat)):
        HBLON = lons[i]
        HBLAT = lats[j]

        HBlons.append(HBLON)
        HBlats.append(HBLAT)

        geometry.append(
            Polygon(
                [
                    (HBLON - delta / 2, HBLAT + delta / 2),
                    (HBLON + delta / 2, HBLAT + delta / 2),
                    (HBLON + delta / 2, HBLAT - delta / 2),
                    (HBLON - delta / 2, HBLAT - delta / 2),
                ]
            )
        )

        try:
            values.append(float(dataset[data_name].isel(lon=i, lat=j).data))
        except Exception:
            values.append(float(dataset[data_name].isel(longitude=i, latitude=j).data))

    data = gpd.GeoDataFrame(crs="EPSG:4326", geometry=geometry)
    data["HBlon"] = HBlons
    data["HBlat"] = HBlats
    data[data_name] = values
    # data['geometry']=geometry
    return data


def gen_grids(bbox, resolution, offset):
    lx = bbox[0]
    rx = bbox[2]
    LLON = round(
        int(lx)
        + resolution * int((lx - int(lx)) / resolution + 0.5)
        + offset
        * (int(lx * 10) / 10 + offset - lx)
        / abs(int(lx * 10) / 10 + offset - lx),
        3,
    )
    RLON = round(
        int(rx)
        + resolution * int((rx - int(rx)) / resolution + 0.5)
        - offset
        * (int(rx * 10) / 10 + offset - rx)
        / abs(int(rx * 10) / 10 + offset - rx + 0.00001),
        3,
    )
    by = bbox[1]
    ty = bbox[3]
    BLAT = round(
        int(by)
        + resolution * int((by - int(by)) / resolution + 0.5)
        + offset
        * (int(by * 10) / 10 + offset - by)
        / abs(int(by * 10) / 10 + offset - by),
        3,
    )
    TLAT = round(
        int(ty)
        + resolution * int((ty - int(ty)) / resolution + 0.5)
        - offset
        * (int(ty * 10) / 10 + offset - ty)
        / abs(int(ty * 10) / 10 + offset - ty),
        3,
    )
    xsize = round((RLON - LLON) / resolution) + 1
    ysize = round((TLAT - BLAT) / resolution) + 1
    lons = np.linspace(LLON, RLON, xsize)
    lats = np.linspace(TLAT, BLAT, ysize)
    geometry = []
    HBlons = []
    HBlats = []
    for i in range(xsize):
        for j in range(ysize):
            HBLON = lons[i]
            HBLAT = lats[j]
            HBlons.append(HBLON)
            HBlats.append(HBLAT)
            geometry.append(
                Polygon(
                    [
                        (
                            round(HBLON - resolution / 2, 3),
                            round(HBLAT + resolution / 2, 3),
                        ),
                        (
                            round(HBLON + resolution / 2, 3),
                            round(HBLAT + resolution / 2, 3),
                        ),
                        (
                            round(HBLON + resolution / 2, 3),
                            round(HBLAT - resolution / 2, 3),
                        ),
                        (
                            round(HBLON - resolution / 2, 3),
                            round(HBLAT - resolution / 2, 3),
                        ),
                    ]
                )
            )
    data = gpd.GeoDataFrame(crs="EPSG:4326", geometry=geometry)
    data["lon"] = HBlons
    data["lat"] = HBlats
    return data


def get_para(data_name):
    if data_name.lower() in ["era5_land"]:
        return 0.1, 0
    elif data_name.lower() in ["gpm"]:
        return 0.1, 0.05
    elif data_name.lower() in ["gfs", "era5"]:
        return 0.25, 0
    else:
        raise Exception("未支持的数据产品")


def gen_mask(watershed, dataname):
    """
    计算流域平均
    Args:
        watershed (GeoDataframe): 必选，流域的矢量数据，通过geopandas读取
        dataname (DataArray): 必选，表示流域mask数据名称
    Returns
        data (Dataframe): 流域编号和对应的平均值
    """
    for index, row in watershed.iterrows():
        # wid = row[filedname]
        # wid = basin_id
        geo = row["geometry"]
        bbox = geo.bounds
        res, offset = get_para(dataname)
        grid = gen_grids(bbox, res, offset)
        grid = grid.to_crs(epsg=3857)
        grid["GRID_AREA"] = grid.area
        grid = grid.to_crs(epsg=4326)
        gs = gpd.GeoSeries.from_wkt([geo.wkt])
        sub = gpd.GeoDataFrame(crs="EPSG:4326", geometry=gs)
        intersects = gpd.overlay(grid, sub, how="intersection")
        intersects = intersects.to_crs(epsg=3857)
        intersects["BASIN_AREA"] = intersects.area
        intersects = intersects.to_crs(epsg=4326)
        intersects["w"] = intersects["BASIN_AREA"] / intersects["GRID_AREA"]
        grids = grid.set_index(["lon", "lat"]).join(
            intersects.set_index(["lon", "lat"]), lsuffix="_left", rsuffix="_right"
        )
        grids = grids.loc[:, ["w"]]
        grids.loc[grids.w.isnull(), "w"] = 0
        wds = grids.to_xarray()
        # wds.to_netcdf(os.path.join(save_dir, f"mask-{wid}-{dataname}.nc"))
        return wds


def gen_mask_smap(smap_cell_array, basin_gdf):
    poly_list = []
    for i in range(0, smap_cell_array.shape[0] - 1):
        for j in range(0, smap_cell_array.shape[1] - 1):
            lon_half = (smap_cell_array[i][j + 1][0] - smap_cell_array[i][j][0]) / 2
            lat_half = (smap_cell_array[i][j][1] - smap_cell_array[i + 1][j][1]) / 2
            corner1 = (smap_cell_array[i][j][0] - lon_half, smap_cell_array[i][j][1] - lat_half)
            corner2 = (smap_cell_array[i][j][0] + lon_half, smap_cell_array[i][j][1] - lat_half)
            corner3 = (smap_cell_array[i][j][0] - lon_half, smap_cell_array[i][j][1] + lat_half)
            corner4 = (smap_cell_array[i][j][0] + lon_half, smap_cell_array[i][j][1] + lat_half)
            geom = Polygon((corner1, corner2, corner4, corner3, corner1))
            poly_list.append(geom)
    grid_gdf = GeoDataFrame(geometry=poly_list)
    # intersects = gpd.overlay(grid_gdf, basin_gdf, how="intersection")
    intersects = gpd.sjoin(grid_gdf, basin_gdf, how="inner")
    intersects["w"] = intersects.area / grid_gdf.geometry.area
    if len(intersects) < len(grid_gdf):
        zero_arr = np.append(np.array([intersects.iloc[len(intersects)-1].to_numpy()[:-2]], dtype=object), [np.nan, 0])
        concat_df = pd.DataFrame(np.repeat(zero_arr, len(grid_gdf)-len(intersects)).
                                 reshape((len(zero_arr), len(grid_gdf)-len(intersects))).T, columns=intersects.columns)
        intersects = pd.concat([intersects, concat_df], ignore_index=True)
    wds = intersects.drop(columns=['AREA', 'index_right', 'geometry']).to_xarray()
    return wds


def gen_single_mask(basin_id, watershed, dataname):
    mask_file_name = f"mask-{basin_id.rstrip('.zip')}-{dataname}.nc"
    s3_mask_path = f"s3://basins-origin/hour_data/1h/grid_data/{mask_file_name}"
    if hdscc.FS.exists(s3_mask_path):
        mask = xr.open_dataset(hdscc.FS.open(s3_mask_path))
    else:
        if dataname in ['gpm', 'gfs', 'era5_land', 'era5']:
            mask = gen_mask(watershed, dataname)
        elif dataname == 'smap':
            # w,e,n,s
            bounds_array = watershed.bounds.to_numpy()[0]
            bbox = [bounds_array[0], bounds_array[2], bounds_array[3], bounds_array[1]]
            # 不管什么时间，同一个bbox下，得到的smap数据和网格总是一致
            lon_array = np.load('smap_lon.npy')
            lat_array = np.load('smap_lat.npy')
            w_index = np.argwhere(lon_array >= bbox[0])[0][0]
            e_index = np.argwhere(lon_array <= bbox[1])[-1][0]
            n_index = np.argwhere(lat_array <= bbox[2])[0][0]
            s_index = np.argwhere(lat_array >= bbox[3])[-1][0]
            lon_slice = lon_array[w_index: e_index + 2]
            lat_slice = lat_array[n_index: s_index + 2]
            smap_cell_array = np.ndarray((lat_slice.shape[0], lon_slice.shape[0]), dtype=object)
            for lat in range(0, lat_slice.shape[0]):
                for lon in range(0, lon_slice.shape[0]):
                    smap_cell_array[lat, lon] = (lon_slice[lon], lat_slice[lat])
            mask = gen_mask_smap(smap_cell_array, watershed)
        else:
            mask = xr.Dataset()
        hdscc.FS.write_bytes(s3_mask_path, mask.to_netcdf())
    return mask


def mean_by_mask(src, var, mask):
    """
    计算流域平均

    Args:
        src (dataset): 必选，流域的网格数据，通过xarray读取
        var (str): 必选，表示网格数据中需要计算的变量名称
        mask (dataset): 必选，表示流域mask，通过xarray读取

    Returns
        m (float): 平均值
    """
    src_array = src[var].to_numpy()
    if 'tp' in src.data_vars:
        mask_array = mask["w"].to_numpy().T
        mask_array_expand = np.expand_dims(mask_array, 0).repeat(src_array.shape[0], 0)
    elif 'sm_surface' in src.data_vars:
        mask_array = mask["w"].to_numpy().reshape(src_array.shape)
        mask_array_expand = np.expand_dims(mask_array, 0)
    else:
        mask_array = mask["w"].to_numpy()
        mask_array_expand = np.expand_dims(mask_array, 0).repeat(src_array.shape[0], 0)
    # src_array = da.from_array(src_array, chunks='auto')
    # mask_array = da.from_array(mask_array, chunks='auto')
    s = np.multiply(mask_array_expand, src_array)
    return np.nansum(s, axis=(1, 2)) / np.sum(mask_array)



import fsspec
import xarray as xr
import glob
import hydrodatasource.configs.config as hdscc
import numpy as np
import geopandas as gpd
from hydrodatasource.configs.config import FS


def test_merge_era5():
    all_daily_dirs = hdscc.FS.glob('s3://era5-origin/era5/grib/single_levels_tp/**', maxdepth=3)
    daily_dirs = [daily for daily in all_daily_dirs if len(daily.split('/')) == 7]
    for dir in daily_dirs:
        date_nlist = dir.split('/')
        year = date_nlist[-3]
        month = date_nlist[-2]
        day = date_nlist[-1]
        daily_files = hdscc.FS.glob(dir + '/**')[1:]
        daily_ds_list = []
        for dfile in daily_files:
            grib_s3file = f'simplecache::s3://{dfile}'
            time_ds = xr.open_dataset(fsspec.open_local(grib_s3file, s3=hdscc.MINIO_PARAM, filecache=
            {'cache_storage': '/tmp/files'}), engine='cfgrib')
            daily_ds_list.append(time_ds)
        daily_ds = xr.concat(daily_ds_list, 'valid_time')
        hdscc.FS.write_bytes(path=f's3://era5-origin/era5/grib/single_levels_tp_daily/{year}/{month}/{day}.nc',
                             value=daily_ds.to_netcdf())


def test_merge_era5l_camels_sl_sd():
    import pandas as pd
    mistake_list = ['10901810', '11405810', '20101940', '20302520']
    old_era5l_ds_15_23 = xr.open_dataset("/ftproot/era5land/563_basins_era5_land_2015_2023.nc")
    old_era5l_ds_15_23 = old_era5l_ds_15_23.astype(np.float32)
    old_era5l_ds_15_23['time_start'] = pd.to_datetime(old_era5l_ds_15_23['time_start'])
    old_era5l_ds_15_23 = old_era5l_ds_15_23.drop_sel(basin_id=mistake_list)
    old_era5l_ds_00_14 = xr.open_dataset("/ftproot/camels_songliao_era5land_2000_2014.nc")
    old_era5l_ds_00_14 = old_era5l_ds_00_14.drop_sel(basin_id=mistake_list).drop('Unnamed: 0')
    old_era5l_ds = xr.concat([old_era5l_ds_00_14, old_era5l_ds_15_23], dim='time_start')
    old_era5l_ds = old_era5l_ds.astype(np.float32)
    sd_ds = xr.open_dataset("/ftproot/songliao_shandong_basins_era5land.nc")
    new_era5l_ds = xr.concat([old_era5l_ds, sd_ds], dim='basin_id')
    new_era5l_ds = new_era5l_ds.astype(np.float32)
    # 不要从自己的工作文件夹中转，直接生成到相应文件夹即可，以防cp命令改变数据
    new_era5l_ds.to_netcdf('/ftproot/632_basins_era5land_fixed.nc')


def test_merge_new_streamflow():
    import pandas as pd
    new_dfs = glob.glob('/ftproot/basins-origin/basins-streamflow-BSAD/basins-streamflow-with BSAD/*.csv',
                        recursive=True)
    basins_shp = gpd.read_file(hdscc.FS.open('s3://basins-origin/basins_shp.zip'))
    rsvr_dfs = []
    for rsvr_file in new_dfs:
        rsvr_part_df = pd.read_csv(rsvr_file, engine='c', parse_dates=['TM'])
        if 'rsvr' in rsvr_file:
            rsvr_part_df['STCD'] = rsvr_part_df['STCD'].astype(str)
            rsvr_part_df = rsvr_part_df.rename(columns={'INQ': 'Q'})
            rsvr_part_df = rsvr_part_df[['STCD', 'TM', 'Q']]
            stcd_area = get_area_from_stcd(basins_shp, rsvr_part_df['STCD'].to_list()[0]).to_list()[0]
            rsvr_part_df['Q'] = rsvr_part_df['Q'].apply(lambda x: x / stcd_area * 3.6)
            rsvr_dfs.append(rsvr_part_df)
        else:
            rsvr_part_df = rsvr_part_df.rename(columns={'INQ': 'Q'})
            rsvr_part_df['STCD'] = np.repeat(rsvr_file.split('/')[-1].split('.')[0], len(rsvr_part_df))
            stcd_area = get_area_from_stcd(basins_shp, rsvr_part_df['STCD'].to_list()[0]).to_list()[0]
            rsvr_part_df['Q'] = rsvr_part_df['Q'].apply(lambda x: x / stcd_area * 3.6)
            rsvr_part_df = rsvr_part_df[['STCD', 'TM', 'Q']]
            rsvr_dfs.append(rsvr_part_df)
    rsvr_df = pd.concat(rsvr_dfs).rename(columns={'TM': 'time_start', 'STCD': 'basin_id', 'Q': 'streamflow'})
    rsvr_df = rsvr_df[(rsvr_df['time_start'] >= pd.to_datetime('2000-01-01 00:00:00')) & (
            rsvr_df['time_start'] <= pd.to_datetime('2023-12-31 23:00:00'))]
    rsvr_df['basin_id'] = rsvr_df['basin_id'].astype(str)
    rsvr_df = rsvr_df.set_index(['time_start', 'basin_id'])
    # 在这里xarray可能会莫名生成几个分钟数据，造成索引错误
    rsvr_ds = xr.Dataset.from_dataframe(rsvr_df)
    time_start_arr = rsvr_ds['time_start'].to_numpy()
    hourly_dr = pd.date_range(time_start_arr[0], time_start_arr[-1], freq='1h')
    rsvr_ds = rsvr_ds.sel(time_start=hourly_dr)
    # rsvr_ds.coords['basin_id'] = rsvr_df['basin_id'].to_numpy()
    rsvr_ds.to_netcdf("/ftproot/100_basins_streamflow.nc")


def get_area_from_stcd(basins_shp, stcd: str):
    geom_stcd = basins_shp.geometry[basins_shp['BASIN_ID'] == stcd]
    return geom_stcd.area * 12345.6789


def test_compare_basin_amount():
    total_shps = gpd.read_file(FS.open('s3://basins-origin/basins_shp.zip'))
    total_basin_ids = total_shps['BASIN_ID'].to_numpy()
    usa_stream_basins = xr.open_dataset("/ftproot/516_basins_2000_2014.nc")['basin_id'].to_numpy()
    chn_stream_basins = xr.open_dataset("/ftproot/100_basins_streamflow.nc")['basin_id'].to_numpy()
    calced_basins = np.concatenate([usa_stream_basins, chn_stream_basins])
    result_diff = np.setdiff1d(total_basin_ids, calced_basins)
    print(result_diff)
    return result_diff


def test_merge_616_basins_streamflow():
    usa_basins_53_15_23 = xr.open_dataset("/ftproot/53_usa_basins_streamflow.nc")
    usa_basins_463_15_23 = xr.open_dataset("/ftproot/usgs_datas_462_basins_after_2019/463_basins_2015_2023.nc")
    usa_basins_516_00_14 = xr.open_dataset("/ftproot/516_basins_2000_2014.nc")
    chn_basins_100_00_23 = xr.open_dataset("/ftproot/100_basins_streamflow.nc")
    chn_basins_100_00_23['time_start'] = chn_basins_100_00_23['time_start'] - np.timedelta64(8, 'h')
    usa_basins_516_15_23 = xr.concat([usa_basins_463_15_23, usa_basins_53_15_23], dim='basin_id')
    usa_basins = xr.concat([usa_basins_516_00_14, usa_basins_516_15_23], dim='time_start')
    basins_ds = xr.concat([usa_basins.drop_duplicates('time_start'), chn_basins_100_00_23], dim='basin_id')
    basins_ds = basins_ds.astype(np.float32)
    basins_ds.coords['basin_id'] = basins_ds['basin_id'].to_numpy()
    basins_ds.to_netcdf('/ftproot/616_basins_streamflow_2000_2023.nc')



"""
Author: Wenyu Ouyang
Date: 2023-10-28 08:28:37
LastEditTime: 2024-03-28 08:41:56
LastEditors: Wenyu Ouyang
Description: Test funcs for processor
FilePath: \hydrodatasource\tests\test_processor.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import geopandas as gpd
from hydrodatasource.processor.minio_process import GeoProcessor, geojson_to_shp


def test_geojson_to_shp(tmp_path):
    # create a temporary GeoJSON file
    input_geojson = os.path.join(tmp_path, "test.geojson")
    points = gpd.points_from_xy([0, 1], [0, 1])
    gdf = gpd.GeoDataFrame({"col1": [1, 2], "geometry": points})
    gdf.to_file(input_geojson, driver="GeoJSON")

    # call the function with the temporary GeoJSON file
    output_folder = os.path.join(tmp_path, "shp_output")
    geojson_to_shp(input_geojson, output_folder=output_folder)

    # check that the Shapefile was created
    assert os.path.exists(os.path.join(output_folder, "test.shp"))

    # check that the ZIP file was created
    assert os.path.exists(os.path.join(tmp_path, "test.zip"))


def test_upload_and_read_shp(tmp_path, minio_paras):
    # create a temporary GeoJSON file
    input_geojson = os.path.join(tmp_path, "test.geojson")
    points = gpd.points_from_xy([0, 1], [0, 1])
    gdf = gpd.GeoDataFrame({"col1": [1, 2], "geometry": points})
    gdf.to_file(input_geojson, driver="GeoJSON")
    geo_processor = GeoProcessor(minio_paras)
    geo_processor.upload_geojson(gj_local_path=input_geojson, gj_mo_name="test.geojson")
    gdf_rd = geo_processor.read_shp("test.zip")
    assert gdf_rd.equals(gdf)



"""
Author: Jianfeng Zhu
Date: 2023-10-22 16:47:36
LastEditTime: 2024-06-25 11:47:47
LastEditors: Wenyu Ouyang
Description: Reader class for grid data in minio
FilePath: \hydrodatasource\hydrodatasource\reader\minio.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import os
import numpy as np
import xarray as xr
from datetime import datetime
import calendar
import dask
import json
import geopandas as gpd

from ..configs.config import FS, MINIO_PARAM, RO
from ..utils.utils import regen_box, creatspinc

dask.config.set({"array.slicing.split_large_chunks": False})


class ERA5LReader:
    """
    用于从minio中读取era5-land数据

    Methods:
        open_dataset(data_variables, start_time, end_time, dataset, bbox): 从minio中读取era5-land数据
        from_shp(data_variables, start_time, end_time, dataset, shp): 通过已有的矢量数据范围从minio服务器读取era5-land数据
        from_aoi(data_variables, start_time, end_time, dataset, aoi): 用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取era5-land数据
        to_netcdf(data_variables, start_time, end_time, dataset, shp, resolution, save_file): 读取数据并保存为本地nc文件
    """

    def __init__(self):
        self._variables = [
            "10 metre U wind component",
            "10 metre V wind component",
            "2 metre dewpoint temperature",
            "2 metre temperature",
            "Evaporation",
            "Evaporation from bare soil",
            "Evaporation from open water surfaces excluding oceans",
            "Evaporation from the top of canopy",
            "Evaporation from vegetation transpiration",
            "Forecast albedo",
            "Lake bottom temperature",
            "Lake ice total depth",
            "Lake ice surface temperature",
            "Lake mix-layer depth",
            "Lake mix-layer temperature",
            "Lake shape factor",
            "Lake total layer temperature",
            "Leaf area index, high vegetation",
            "Leaf area index, low vegetation",
            "Potential evaporation",
            "Runoff",
            "Skin reservoir content",
            "Skin temperature",
            "Snow albedo",
            "Snow cover",
            "Snow density",
            "Snow depth",
            "Snow depth water equivalent",
            "Snow evaporation",
            "Snowfall",
            "Snowmelt",
            "Soil temperature level 1",
            "Soil temperature level 2",
            "Soil temperature level 3",
            "Soil temperature level 4",
            "Sub-surface runoff",
            "Surface latent heat flux",
            "Surface net solar radiation",
            "Surface net thermal radiation",
            "Surface pressure",
            "Surface runoff",
            "Surface sensible heat flux",
            "Surface solar radiation downwards",
            "Surface thermal radiation downwards",
            "Temperature of snow layer",
            "Total precipitation",
            "Volumetric soil water layer 1",
            "Volumetric soil water layer 2",
            "Volumetric soil water layer 3",
            "Volumetric soil water layer 4",
        ]

        self._accumulated = [
            "Evaporation",
            "Evaporation from bare soil",
            "Evaporation from open water surfaces excluding oceans",
            "Evaporation from the top of canopy",
            "Evaporation from vegetation transpiration",
            "Potential evaporation",
            "Runoff",
            "Snow evaporation",
            "Snowfall",
            "Snowmelt",
            "Sub-surface runoff",
            "Surface latent heat flux",
            "Surface net solar radiation",
            "Surface net thermal radiation",
            "Surface runoff",
            "Surface sensible heat flux",
            "Surface solar radiation downwards",
            "Surface thermal radiation downwards",
            "Total precipitation",
        ]
        self._bucket_name = "test"

    def open_dataset(
        self,
        data_variables=["Total precipitation"],
        start_time=None,
        end_time=None,
        dataset="wis",
        bbox=None,
        time_chunks=24,
    ):
        """
        从minio服务器读取era5-land数据

        Args:
            data_variables (list): 数据变量列表
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            bbox (list|tuple): 四至范围
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        if end_time <= start_time:
            raise Exception("结束时间不能早于开始时间")

        if bbox[0] > bbox[2] or bbox[1] > bbox[3]:
            raise Exception("四至范围格式错误")

        if dataset != "wis" and dataset != "camels":
            raise Exception("dataset参数错误")

        if dataset == "wis":
            self._dataset = "geodata"
        elif dataset == "camels":
            self._dataset = "camdata"

        with FS.open(
            os.path.join(self._bucket_name, f"{self._dataset}/era5_land/era5l.json")
        ) as f:
            cont = json.load(f)
            self._starttime = np.datetime64(cont["start"])
            self._endtime = np.datetime64(cont["end"])
            self._bbox = cont["bbox"]

        chunks = {"time": time_chunks}
        ds = xr.open_dataset(
            "reference://",
            engine="zarr",
            chunks=chunks,
            backend_kwargs={
                "consolidated": False,
                "storage_options": {
                    # no matter you run code in windows or linux, the bucket's format should be Linux style
                    # so we don't use os.join.path
                    "fo": f"s3://{self._bucket_name}/{self._dataset}/era5_land/era5_land_.json",
                    "target_protocol": "s3",
                    "target_options": RO,
                    "remote_protocol": "s3",
                    "remote_options": RO,
                },
            },
        )

        ds = ds.filter_by_attrs(long_name=lambda v: v in data_variables)
        ds = ds.rename({"longitude": "lon", "latitude": "lat"})
        ds = ds.transpose("time", "lon", "lat")

        start_time = max(start_time, self._starttime)
        end_time = min(end_time, self._endtime)
        times = slice(start_time, end_time)
        ds = ds.sel(time=times)

        bbox = regen_box(bbox, 0.1, 0)

        if bbox[0] < self._bbox[0]:
            left = self._bbox[0]
        else:
            left = bbox[0]

        if bbox[1] < self._bbox[1]:
            bottom = self._bbox[1]
        else:
            bottom = bbox[1]

        if bbox[2] > self._bbox[2]:
            right = self._bbox[2]
        else:
            right = bbox[2]

        if bbox[3] > self._bbox[3]:
            top = self._bbox[3]
        else:
            top = bbox[3]

        longitudes = slice(left - 0.00001, right + 0.00001)
        latitudes = slice(bottom - 0.00001, top + 0.00001)

        ds = ds.sortby("lat", ascending=True)
        ds = ds.sel(lon=longitudes, lat=latitudes)

        return ds

    def from_shp(
        self,
        data_variables=["Total precipitation"],
        start_time=None,
        end_time=None,
        dataset="wis",
        shp=None,
        time_chunks=24,
    ):
        """
        通过已有的矢量数据范围从minio服务器读取era5-land数据

        Args:
            data_variables (list): 数据变量列表
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            shp (str): 矢量数据路径
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        gdf = gpd.GeoDataFrame.from_file(shp)
        b = gdf.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0,
        )

        ds = self.open_dataset(
            data_variables, start_time, end_time, dataset, bbox, time_chunks
        )

        return self.open_dataset(
            data_variables, start_time, end_time, dataset, bbox, time_chunks
        )

    def from_aoi(
        self,
        data_variables=["Total precipitation"],
        start_time=None,
        end_time=None,
        dataset="wis",
        aoi: gpd.GeoDataFrame = None,
        time_chunks=24,
    ):
        """
        用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取era5-land数据

        Args:
            data_variables (list): 数据变量列表
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        b = aoi.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0,
        )

        ds = self.open_dataset(
            data_variables, start_time, end_time, dataset, bbox, time_chunks
        )

        return ds

    def to_netcdf(
        self,
        data_variables=["Total precipitation"],
        start_time=None,
        end_time=None,
        dataset="wis",
        shp=None,
        resolution="hourly",
        save_file="era5.nc",
        time_chunks=24,
    ):
        """
        读取数据并保存为本地nc文件

        Args:
            data_variables (list): 数据变量列表
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            shp (str): 已有的矢量数据路径
            resolution (str): 输出的时间分辨率
            save_file (str): 输出的文件路径
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        gdf = gpd.GeoDataFrame.from_file(shp)
        b = gdf.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0,
        )

        if resolution == "hourly":
            ds = self.open_dataset(
                data_variables, start_time, end_time, dataset, bbox, time_chunks
            )

            if ds.to_netcdf(save_file) == None:
                print(save_file, "已生成")
                ds = xr.open_dataset(save_file)
                return ds

        if resolution == "daily":
            start_time = np.datetime64(f"{str(start_time)[:10]}T01:00:00.000000000")
            end_time = np.datetime64(str(end_time)[:10]) + 1
            end_time = np.datetime64(f"{str(end_time)}T00:00:00.000000000")

            ds = self.open_dataset(
                data_variables, start_time, end_time, dataset, bbox, time_chunks
            )

            days = ds["time"].size // 24

            data_vars = {}
            for k, v in ds.data_vars.items():
                data_vars[k] = v.attrs

            daily_arr = []

            for var, attr in data_vars.items():
                a = ds[var].to_numpy()

                if attr["long_name"] in self._accumulated:
                    xlist = [x for x in range(a.shape[0]) if x % 24 != 23]
                    _a = np.delete(a, xlist, axis=0)

                    daily_arr.append(_a)

                else:
                    r = np.split(a, days, axis=0)
                    _r = [
                        np.expand_dims(np.mean(r[i], axis=0), axis=0)
                        for i in range(len(r))
                    ]
                    __r = np.concatenate(_r)

                    daily_arr.append(__r)

            lats = ds["lat"].to_numpy()
            lons = ds["lon"].to_numpy()

            start_time = np.datetime64(str(start_time)[:10])

            creatspinc(daily_arr, data_vars, lats, lons, start_time, save_file, "daily")

            new = xr.open_dataset(save_file)
            print(save_file, "已生成")
            return new

        if resolution == "6-hourly":
            start_time = np.datetime64(f"{str(start_time)[:10]}T01:00:00.000000000")
            end_time = np.datetime64(str(end_time)[:10]) + 1
            end_time = np.datetime64(f"{str(end_time)}T00:00:00.000000000")

            ds = self.open_dataset(
                data_variables, start_time, end_time, dataset, bbox, time_chunks
            )

            days = ds["time"].size // 6

            data_vars = {}
            for k, v in ds.data_vars.items():
                data_vars[k] = v.attrs

            daily_arr = []

            for var, attr in data_vars.items():
                a = ds[var].to_numpy()

                if attr["long_name"] in self._accumulated:
                    xlist = [x for x in range(a.shape[0]) if x % 6 != 5]
                    _a = np.delete(a, xlist, axis=0)

                    daily_arr.append(_a)

                else:
                    r = np.split(a, days, axis=0)
                    _r = [
                        np.expand_dims(np.mean(r[i], axis=0), axis=0)
                        for i in range(len(r))
                    ]
                    __r = np.concatenate(_r)

                    daily_arr.append(__r)

            lats = ds["lat"].to_numpy()
            lons = ds["lon"].to_numpy()

            # start_time = np.datetime64(f'{str(start_time)[:10]}')
            year = int(f"{str(start_time)[0:4]}")
            month = int(f"{str(start_time)[5:7]}")
            day = int(f"{str(start_time)[8:10]}")
            dt = datetime(year, month, day, 0, 0, 0)

            creatspinc(daily_arr, data_vars, lats, lons, dt, save_file, "6-hourly")

            new = xr.open_dataset(save_file)
            print(save_file, "已生成")
            return new


class GPMReader:
    """
    用于从minio中读取gpm数据

    Methods:
        open_dataset(start_time, end_time, dataset, bbox, time_resolution): 从minio中读取gpm数据
        from_shp(start_time, end_time, dataset, shp, time_resolution): 通过已有的矢量数据范围从minio服务器读取gpm数据
        from_aoi(start_time, end_time, dataset, aoi, time_resolution): 用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gpm数据
    """

    def __init__(self):
        self._bucket_name = "test"

    def _get_dataset(self, scale, start_time, end_time, bbox, time_chunks):
        year = str(start_time)[:4]
        month = str(start_time)[5:7].zfill(2)
        day = str(self._endtime)[8:10].zfill(2)

        if scale == "Y":
            minio_path = f"s3://{self._bucket_name}/{self._dataset}/gpm{self._time_resolution}/{year}/gpm{year}_inc.json"

        elif scale == "M":
            minio_path = f"s3://{self._bucket_name}/{self._dataset}/gpm{self._time_resolution}/{year}/{month}/gpm{year}{month}_inc.json"

        chunks = {"time": time_chunks}
        ds = xr.open_dataset(
            "reference://",
            engine="zarr",
            chunks=chunks,
            backend_kwargs={
                "consolidated": False,
                "storage_options": {
                    "fo": minio_path,
                    "target_protocol": "s3",
                    "target_options": RO,
                    "remote_protocol": "s3",
                    "remote_options": RO,
                },
            },
        )

        # if self._time_resolution == '1d':
        #     ds = cf2datetime(ds)

        ds = ds["precipitationCal"]
        # ds.to_dataframe().filter(['precipitationCal','precipitationUncal']).to_xarray()

        # ds = ds.rename({"longitude": "lon", "latitude": "lat"})
        ds = ds.transpose("time", "lon", "lat")

        times = slice(start_time, end_time)
        ds = ds.sel(time=times)

        left = bbox[0]
        right = bbox[2]
        bottom = bbox[1]
        top = bbox[3]

        longitudes = slice(left - 0.00001, right + 0.00001)
        latitudes = slice(bottom - 0.00001, top + 0.00001)

        ds = ds.sortby("lat", ascending=True)
        ds = ds.sel(lon=longitudes, lat=latitudes)

        return ds

    def open_dataset(
        self,
        start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
        end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
        dataset="wis",
        bbox=(121, 39, 122, 40),
        time_resolution="1d",
        time_chunks=48,
    ):
        """
        从minio服务器读取gpm数据

        Args:
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            bbox (list|tuple): 四至范围
            time_resolution (str): 1d或30m
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        if end_time <= start_time:
            raise Exception("结束时间不能早于开始时间")

        if bbox[0] > bbox[2] or bbox[1] > bbox[3]:
            raise Exception("四至范围错误")

        if dataset != "wis" and dataset != "camels":
            raise Exception("dataset参数错误")

        if time_resolution != "1d" and time_resolution != "30m":
            raise Exception("time_resolution参数错误")

        # dataset_name = get_dataset_name()
        if dataset == "wis":
            self._dataset = "geodata"
        elif dataset == "camels":
            self._dataset = "camdata"

        if time_resolution == "1d":
            self._time_resolution = "1d"
        elif time_resolution == "30m":
            self._time_resolution = ""

        with FS.open(
            os.path.join(
                self._bucket_name,
                f"{self._dataset}/gpm{self._time_resolution}/gpm{self._time_resolution}.json",
            )
        ) as f:
            cont = json.load(f)
            self._starttime = np.datetime64(cont["start"])
            self._endtime = np.datetime64(cont["end"])
            self._bbox = cont["bbox"]

        if start_time < self._starttime:
            start_time = self._starttime

        if end_time > self._endtime:
            end_time = self._endtime

        bbox = regen_box(bbox, 0.1, 0.05)

        if bbox[0] < self._bbox[0]:
            bbox[0] = self._bbox[0]
        if bbox[1] < self._bbox[1]:
            bbox[1] = self._bbox[1]
        if bbox[2] > self._bbox[2]:
            bbox[2] = self._bbox[2]
        if bbox[3] > self._bbox[3]:
            bbox[3] = self._bbox[3]

        year_start = int(str(start_time)[:4])
        year_end = int(str(end_time)[:4])
        end_year = int(str(self._endtime)[:4])

        if year_end < end_year:
            if year_start == year_end:
                ds = self._get_dataset(
                    scale="Y",
                    start_time=start_time,
                    end_time=end_time,
                    bbox=bbox,
                    time_chunks=time_chunks,
                )
                return ds

            elif year_start < year_end:
                dss = []
                years = range(year_start, year_end + 1)
                for year in years:
                    if year == year_start:
                        dss.append(
                            self._get_dataset(
                                scale="Y",
                                start_time=start_time,
                                end_time=np.datetime64(
                                    f"{year}-12-31T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

                    elif year == year_end:
                        dss.append(
                            self._get_dataset(
                                scale="Y",
                                start_time=np.datetime64(
                                    f"{year}-01-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

                    else:
                        dss.append(
                            self._get_dataset(
                                scale="Y",
                                start_time=np.datetime64(
                                    f"{year}-01-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{year}-12-31T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                return xr.merge(dss)

        else:
            month_end = int(str(end_time)[5:7])
            end_month = int(str(self._endtime)[5:7])

            if year_start == year_end:
                month_start = int(str(start_time)[5:7])
                if month_start == month_end:
                    return self._get_dataset(
                        scale="M",
                        start_time=start_time,
                        end_time=end_time,
                        bbox=bbox,
                        time_chunks=time_chunks,
                    )
                dss = []
                for m in range(month_start, month_end + 1):
                    if m == month_start:
                        d = calendar.monthrange(year_start, m)[1]
                        dss.append(
                            self._get_dataset(
                                scale="M",
                                start_time=start_time,
                                end_time=np.datetime64(
                                    f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    elif m == month_end:
                        dss.append(
                            self._get_dataset(
                                scale="M",
                                start_time=np.datetime64(
                                    f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=end_time,
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )
                    else:
                        d = calendar.monthrange(year_start, m)[1]
                        dss.append(
                            self._get_dataset(
                                scale="M",
                                start_time=np.datetime64(
                                    f"{year_start}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(
                                    f"{year_start}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                ),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

            else:
                dss = []

                for y in range(year_start, year_end + 1):
                    if y == year_start:
                        dss.append(
                            self._get_dataset(
                                scale="Y",
                                start_time=start_time,
                                end_time=np.datetime64(f"{y}-12-31T23:30:00.000000000"),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

                    elif y == year_end:
                        for m in range(1, month_end + 1):
                            if m == month_end:
                                dss.append(
                                    self._get_dataset(
                                        scale="M",
                                        start_time=np.datetime64(
                                            f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                        ),
                                        end_time=end_time,
                                        bbox=bbox,
                                        time_chunks=time_chunks,
                                    )
                                )
                            else:
                                d = calendar.monthrange(y, m)[1]
                                dss.append(
                                    self._get_dataset(
                                        scale="M",
                                        start_time=np.datetime64(
                                            f"{str(y).zfill(4)}-{str(m).zfill(2)}-01T00:00:00.000000000"
                                        ),
                                        end_time=np.datetime64(
                                            f"{str(y).zfill(4)}-{str(m).zfill(2)}-{str(d).zfill(2)}T23:30:00.000000000"
                                        ),
                                        bbox=bbox,
                                        time_chunks=time_chunks,
                                    )
                                )

                    else:
                        dss.append(
                            self._get_dataset(
                                scale="Y",
                                start_time=np.datetime64(
                                    f"{y}-01-01T00:00:00.000000000"
                                ),
                                end_time=np.datetime64(f"{y}-12-31T23:30:00.000000000"),
                                bbox=bbox,
                                time_chunks=time_chunks,
                            )
                        )

            return xr.merge(dss)

    def from_shp(
        self,
        start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
        end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
        dataset="wis",
        shp=None,
        time_resolution="1d",
        time_chunks=48,
    ):
        """
        通过已有的矢量数据范围从minio服务器读取gpm数据

        Args:
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            shp (str): 矢量数据路径
            time_resolution (str): 1d或30m
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        gdf = gpd.GeoDataFrame.from_file(shp)
        b = gdf.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0.05,
        )

        ds = self.open_dataset(
            start_time, end_time, dataset, bbox, time_resolution, time_chunks
        )
        return ds

    def from_aoi(
        self,
        start_time=np.datetime64("2023-01-01T00:00:00.000000000"),
        end_time=np.datetime64("2023-01-02T00:00:00.000000000"),
        dataset="wis",
        aoi: gpd.GeoDataFrame = None,
        time_resolution="1d",
        time_chunks=48,
    ):
        """
        用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gpm数据

        Args:
            start_time (datetime64): 开始时间
            end_time (datetime64): 结束时间
            dataset (str): wis或camels
            aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
            time_resolution (str): 1d或30m
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        b = aoi.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0.05,
        )
        ds = self.open_dataset(
            start_time, end_time, dataset, bbox, time_resolution, time_chunks
        )
        return ds


class GFSReader:
    """
    用于从minio中读取gpm数据

    Attributes:
        variables (dict): 变量名称及缩写

    Methods:
        open_dataset(data_variables, creation_date, creation_time, bbox): 从minio中读取gfs数据
        from_shp(data_variables, creation_date, creation_time, shp): 通过已有的矢量数据范围从minio服务器读取gfs数据
        from_aoi(data_variables, creation_date, creation_time, aoi): 用过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gfs数据
    """

    def __init__(self):
        self._variables = {
            "dswrf": "downward_shortwave_radiation_flux",
            "pwat": "precipitable_water_entire_atmosphere",
            "2r": "relative_humidity_2m_above_ground",
            "2sh": "specific_humidity_2m_above_ground",
            "2t": "temperature_2m_above_ground",
            "tcc": "total_cloud_cover_entire_atmosphere",
            "tp": "total_precipitation_surface",
            "10u": "u_component_of_wind_10m_above_ground",
            "10v": "v_component_of_wind_10m_above_ground",
        }
        self._bucket_name = "test"
        self._default = "tp"

    @property
    def variables(self):
        return self._variables

    @property
    def default_variable(self):
        return self._default

    def set_default_variable(self, short_name):
        if short_name in self._variables.keys():
            self._default = short_name
        else:
            raise Exception("变量设置错误")

    def open_dataset(
        self,
        creation_date=np.datetime64("2022-09-01"),
        creation_time="00",
        dataset="wis",
        bbox=(115, 38, 136, 54),
        time_chunks=24,
    ):
        """
        从minio服务器读取gfs数据

        Args:
            creation_date (datetime64): 创建日期
            creation_time (datetime64): 创建时间，即00\06\12\18之一
            dataset (str): wis或camels
            bbox (list|tuple): 四至范围
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        if bbox[0] > bbox[2] or bbox[1] > bbox[3]:
            raise Exception("四至范围格式错误")

        if dataset != "wis" and dataset != "camels":
            raise Exception("dataset参数错误")

        if dataset == "wis":
            self._dataset = "geodata"
        elif dataset == "camels":
            self._dataset = "camdata"

        with FS.open(
            os.path.join(self._bucket_name, f"{self._dataset}/gfs/gfs.json")
        ) as f:
            cont = json.load(f)
            self._paras = cont

        short_name = self._default
        full_name = self._variables[short_name]

        start = np.datetime64(self._paras[short_name][0]["start"])
        end = np.datetime64(self._paras[short_name][-1]["end"])

        if creation_date < start or creation_date > end:
            print("超出时间范围！")
            return

        if creation_time not in ["00", "06", "12", "18"]:
            print("creation_time必须是00、06、12、18之一！")
            return

        year = str(creation_date.astype("object").year)
        month = str(creation_date.astype("object").month).zfill(2)
        day = str(creation_date.astype("object").day).zfill(2)

        change = np.datetime64("2022-09-01")
        if creation_date < change:
            json_url = f"s3://{self._bucket_name}/{self._dataset}/gfs/gfs_history/{year}/{month}/{day}/gfs{year}{month}{day}.t{creation_time}z.0p25.json"
        else:
            json_url = f"s3://{self._bucket_name}/{self._dataset}/gfs/{short_name}/{year}/{month}/{day}/gfs{year}{month}{day}.t{creation_time}z.0p25.json"

        chunks = {"valid_time": time_chunks}
        ds = xr.open_dataset(
            "reference://",
            engine="zarr",
            chunks=chunks,
            backend_kwargs={
                "consolidated": False,
                "storage_options": {
                    "fo": json_url,
                    "target_protocol": "s3",
                    "target_options": RO,
                    "remote_protocol": "s3",
                    "remote_options": RO,
                },
            },
        )

        if creation_date < change:
            ds = ds[full_name]
            box = self._paras[short_name][0]["bbox"]
        else:
            box = self.paras[short_name][-1]["bbox"]

        # ds = ds.filter_by_attrs(long_name=lambda v: v in data_variables)
        ds = ds.rename({"longitude": "lon", "latitude": "lat"})
        # ds = ds.transpose('time','valid_time','lon','lat')

        bbox = regen_box(bbox, 0.25, 0)

        if bbox[0] < box[0]:
            left = box[0]
        else:
            left = bbox[0]

        if bbox[1] < box[1]:
            bottom = box[1]
        else:
            bottom = bbox[1]

        if bbox[2] > box[2]:
            right = box[2]
        else:
            right = bbox[2]

        if bbox[3] > box[3]:
            top = box[3]
        else:
            top = bbox[3]

        longitudes = slice(left - 0.00001, right + 0.00001)
        latitudes = slice(bottom - 0.00001, top + 0.00001)

        ds = ds.sortby("lat", ascending=True)
        ds = ds.sel(lon=longitudes, lat=latitudes)

        return ds

    def from_shp(
        self,
        creation_date=np.datetime64("2022-09-01"),
        creation_time="00",
        dataset="wis",
        shp=None,
        time_chunks=24,
    ):
        """
        通过已有的矢量数据范围从minio服务器读取gfs数据

        Args:
            creation_date (datetime64): 创建日期
            creation_time (datetime64): 创建时间，即00\06\12\18之一
            dataset (str): wis或camels
            shp (str): 矢量数据路径
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """

        gdf = gpd.GeoDataFrame.from_file(shp)
        b = gdf.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0,
        )

        ds = self.open_dataset(creation_date, creation_time, dataset, bbox, time_chunks)

        return ds

    def from_aoi(
        self,
        creation_date=np.datetime64("2022-09-01"),
        creation_time="00",
        dataset="wis",
        aoi: gpd.GeoDataFrame = None,
        time_chunks=24,
    ):
        """
        通过已有的GeoPandas.GeoDataFrame对象从minio服务器读取gfs数据

        Args:
            creation_date (datetime64): 创建日期
            creation_time (datetime64): 创建时间，即00\06\12\18之一
            dataset (str): wis或camels
            aoi (GeoDataFrame): 已有的GeoPandas.GeoDataFrame对象
            time_chunks (int): 分块数量

        Returns:
            dataset (Dataset): 读取结果
        """
        b = aoi.bounds
        bbox = regen_box(
            (b.loc[0]["minx"], b.loc[0]["miny"], b.loc[0]["maxx"], b.loc[0]["maxy"]),
            0.1,
            0,
        )

        ds = self.open_dataset(creation_date, creation_time, dataset, bbox, time_chunks)

        return ds



"""
Author: Wenyu Ouyang
Date: 2024-04-08 09:52:32
LastEditTime: 2024-04-09 17:57:09
LastEditors: Wenyu Ouyang
Description: Anomaly process for rainfall data
FilePath: \hydrodatasource\hydrodatasource\cleaner\rain_anomaly.py
Copyright (c) 2023-2024 Wenyu Ouyang. All rights reserved.
"""

import xarray as xr
import pandas as pd
import numpy as np
import os


# 数据极值筛查
def rainfall_extreme_filter(df):
    """
    根据汛期和非汛期的极值阈值筛查DataFrame中的降雨数据，超过极值的数据将其DRP值设置为0。

    参数:
    - df: 包含降雨数据的DataFrame。

    返回:
    - 经过极值筛查处理后的DataFrame。
    """
    # 设置汛期与非汛期极值阈值
    extreme_value_flood = 200
    extreme_value_non_flood = 50
    df["TM"] = pd.to_datetime(df["TM"])
    # 识别汛期
    df["Is_Flood_Season"] = df["TM"].apply(lambda x: 6 <= x.month <= 9)

    # 对超过极值阈值的数据进行处理，将DRP值设置为0
    df.loc[
        (df["Is_Flood_Season"] == True) & (df["DRP"] > extreme_value_flood), "DRP"
    ] = 0
    df.loc[
        (df["Is_Flood_Season"] == False) & (df["DRP"] > extreme_value_non_flood), "DRP"
    ] = 0

    return df


# 数据梯度筛查
def rainfall_gradient_filter(df):
    """
    处理DataFrame中的降雨数据，考虑汛期和非汛期的不同梯度阈值。

    参数:
    - df: 包含降雨数据的DataFrame。

    返回:
    - 处理后的DataFrame。
    """
    # 原始总降雨量
    original_total_rainfall = df["DRP"].sum()

    # 计算降雨量变化梯度
    df["DRP_Change"] = df["DRP"].diff()

    # 汛期与非汛期梯度阈值
    gradient_threshold_flood = 20
    gradient_threshold_non_flood = 10

    # 识别汛期
    df["TM"] = pd.to_datetime(df["TM"])
    df["Is_Flood_Season"] = df["TM"].apply(lambda x: 6 <= x.month <= 9)

    # 处理异常值
    df.loc[
        (df["Is_Flood_Season"] == True)
        & (df["DRP_Change"].abs() > gradient_threshold_flood),
        "DRP",
    ] = 0
    df.loc[
        (df["Is_Flood_Season"] == False)
        & (df["DRP_Change"].abs() > gradient_threshold_non_flood),
        "DRP",
    ] = 0

    # 调整后的总降雨量
    adjusted_total_rainfall = df["DRP"].sum()

    # 打印数据总量的变化
    print(f"Original Total Rainfall: {original_total_rainfall} mm")
    print(f"Adjusted Total Rainfall: {adjusted_total_rainfall} mm")
    print(f"Change: {adjusted_total_rainfall - original_total_rainfall} mm")

    return df


# 空间数据对比


# 数据格式规范
def rainfall_format_normalization(df, freq="h"):
    # 转换时间列为datetime类型
    df["TM"] = pd.to_datetime(df["TM"], format="%Y-%m-%d %H:%M:%S", errors="coerce")
    # 尝试转换STCD列为整数，如果失败，则转换为字符串
    try:
        df["STCD"] = df["STCD"].astype(int).astype(str)
    except ValueError:
        df["STCD"] = df["STCD"].astype(str)

    # 生成完整的时间序列，确保没有间隔
    full_time_series = pd.date_range(start=df["TM"].min(), end=df["TM"].max(), freq="h")
    full_df = pd.DataFrame(full_time_series, columns=["TM"])

    # 确保合并前两个DataFrame的时间列数据类型一致
    df["TM"] = pd.to_datetime(df["TM"])
    full_df["TM"] = pd.to_datetime(full_df["TM"])

    # 合并原始数据到完整的时间序列中
    df_complete = pd.merge(full_df, df, on="TM", how="left")

    # 插补缺失数据
    stcd_fill_value = None if df["STCD"].dropna().empty else df["STCD"].dropna().iloc[0]
    df_complete["DRP"].fillna(0, inplace=True)
    df_complete["STCD"].fillna(stcd_fill_value, inplace=True)

    return df_complete



import geopandas as gpd
import pandas as pd
import pytest
import numpy as np
import xarray as xr
import hydrodatasource.configs.config as conf
from hydrodatasource.reader import access_fs


def test_read_zz_stations_ts():
    # 读取csv文件
    zz_stations = access_fs.spec_path(
        "stations-origin/zz_stations/hour_data/1h/zz_CHN_songliao_10800100.csv",
        head="minio",
    )
    print(zz_stations)



import hydrodatasource.configs.config as conf
import xarray as xr
import numpy as np
import pandas as pd


def test_read_gfs_csv():
    gfs_nc_files = [file for file in conf.FS.glob('s3://grids-origin/GFS/**') if 'nc4' in file]
    bbox_list = []
    start_time = []
    end_list = []
    res_lon_list = []
    res_lat_list = []
    for i in range(0, len(gfs_nc_files)):
        test_ds_i = xr.open_dataset(conf.FS.open(gfs_nc_files[i]))
        bbox = [np.min(test_ds_i['longitude'].to_numpy()), np.max(test_ds_i['longitude'].to_numpy()),
                np.max(test_ds_i['latitude'].to_numpy()), np.min(test_ds_i['latitude'].to_numpy())]
        bbox_list.append(bbox)
        start_time.append(test_ds_i['time'].to_numpy()[0])
        end_list.append(test_ds_i['time'].to_numpy()[-1])
        lon_res = abs(np.diff(test_ds_i['longitude'].to_numpy())[0])
        res_lon_list.append(lon_res)
        lat_res = abs(np.diff(test_ds_i['latitude'].to_numpy())[0])
        res_lat_list.append(lat_res)
    test_pd = pd.DataFrame(
        {'bbox': bbox_list, 'time_start': start_time, 'time_end': end_list, 'res_lon': lon_res, 'res_lat': lat_res,
         'path': ['s3://' + file for file in gfs_nc_files]})
    test_pd.to_csv('gfs_metadata.csv')
